#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)
from jgdv import Mixin, Proto
from jgdv.files.tags import SubstitutionFile

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from . import _interface as MAPI
from bibble.util.middlecore import IdenBlockMiddleware

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from bibtexparser import Library
    type Entry = model.Entry
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--|

@Proto(API.WriteTime_p)
class TagsWriter(IdenBlockMiddleware):
    """
      Reduce tag set to a string.
      Pass in to_keywords=True to convert tags -> keywords for bibtex2html
    """

    def __init__(self, *, to_keywords:bool=False, **kwargs):
        super().__init__(**kwargs)
        self._to_keywords = to_keywords

    def on_write(self):
        Never()

    def transform_Entry(self, entry:Entry, library:Library) -> list[Entry]:
        match entry.get(MAPI.TAGS_K):
            case None:
                self._logger.warning("Entry has No Tags on write: %s", entry.key)
                entry.set_field(model.Field(MAPI.TAGS_K, ""))
            case model.Field(value=val) if not bool(val): # type: ignore
                self._logger.warning("Entry has No Tags on write: %s", entry.key)
                entry.set_field(model.Field(MAPI.TAGS_K, ""))
            case model.Field(value=set() as vals):
                entry.set_field(model.Field(MAPI.TAGS_K, ",".join(sorted(vals)))) # type: ignore

        if self._to_keywords:
            entry.set_field(model.Field(MAPI.KEYWORDS_K, entry.get(MAPI.TAGS_K).value))

        return [entry]



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from pylatexenc.latexencode import (RULE_REGEX, UnicodeToLatexConversionRule,
                                    UnicodeToLatexEncoder)
from pylatexenc.latex2text import (LatexNodes2Text, MacroTextSpec,
                                   get_default_latex_context_db)
from pylatexenc.macrospec import LatexContextDb

from bibtexparser import model

from . import _interface as LAPI
from jgdv import Mixin
# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe, VList
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type U2LRule = UnicodeToLatexConversionRule
    type Entry = model.Field
    type Block = model.Block
    from bibtexparser.library import Library
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class UnicodeHelper_m:
    """
    A Helper for using pylatexenc

    Builds Encoders using lists of str pairs
    and Decoders using dicts
    """

    @staticmethod
    def prep_encode_tuples(tuples:list) -> list:
        """ prep tuples of str,str for conversion rules,
        as re.Pattern,str
        """
        result = []
        for x,y in tuples:
            assert(isinstance(y, str))
            match x:
                case str():
                    result.append((re.compile(x), y))
                case re.Pattern():
                    result.append((x, y))
                case x:
                    raise TypeError(type(x))
        else:
            return result

    @staticmethod
    def build_encode_rule(tuples:list) -> U2LRule:
        compiled = UnicodeHelper_m.prep_encode_tuples(tuples)
        return UnicodeToLatexConversionRule(rule_type=RULE_REGEX, rule=compiled)

    @staticmethod
    def build_decode_rule(pair:tuple) -> MacroTextSpec:
        name, replacement = pair
        return MacroTextSpec("url", simplify_repl="%s")

    @staticmethod
    def build_encoder(*, rules:list[str|U2LRule], kwargs:dict) -> UnicodeToLatexEncoder:
        if any(not isinstance(x, str|UnicodeToLatexConversionRule) for x in rules):
            raise TypeError("Encoding Rules need to be UniicodeToLatexConversionRules, use LatexWriter.build_encode_rule")

        rules.append(LAPI.DEFAULT_RULES_K)
        logging.debug("Building Latex Encoder: %s", rules)
        return UnicodeToLatexEncoder(conversion_rules=rules, **kwargs)

    @staticmethod
    def build_decoder(*, rules:dict[str,VList[MacroTextSpec]], kwargs:dict) -> LatexNodes2Text:
        context_db : LatexContextDb = get_default_latex_context_db()
        logging.debug("Building Latex Decoder: %s", rules)
        for cat, y in rules.items():
            match y:
                case list() if all(isinstance(y2, MacroTextSpec) for y2 in y):
                    context_db.add_context_category(cat, prepend=True, macros=y)
                case MacroTextSpec():
                     context_db.add_context_category(cat, prepend=True, macros=[y])
                case _:
                    raise TypeError("Bad Decode Rules Specified", cat, y)

        return LatexNodes2Text(latex_context=context_db, **kwargs)

    def rebuild_encoder(self, *, rules:Maybe[list[U2LRule]]=None, **kwargs) -> None:
        """ Accumulates rules and rebuilds the encoder """
        self._total_rules += [x for x in (rules or []) if x not in self._total_rules]
        self._total_options.update(kwargs)
        self._encoder = self.build_encoder(rules=self._total_rules[:], kwargs=self._total_options)

    def rebuild_decoder(self, *, rules:dict=None, **kwargs) -> None:
        self._total_rules.update(rules or {})
        self._total_options.update(kwargs)
        self._decoder = self.build_decoder(rules=self._total_rules, kwargs=self._total_options)

    def _test_encode(self, text) -> str:
        """ utility to test latex encoding """
        return self._encoder.unicode_to_latex(text)

    def _test_decode(self, text:str) -> str:
        """ utility to test decoding """
        return self._decoder.latex_to_text(text)



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from jgdv import Proto, Mixin
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)
from jgdv.files.tags import SubstitutionFile

# ##-- end 3rd party imports

import bibble._interface as API
from bibble.util.middlecore import IdenBlockMiddleware

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Entry = model.Entry
    type Field = model.Field
    from bibtexparser.library import Library
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(API.WriteTime_p)
class FieldSorter(IdenBlockMiddleware):
    """ Sort the entries of a field
    firsts are exact matches that go at the front.
    lasts are a list of patterns to match on
    """

    _first_defaults : ClassVar[list[str]] = []
    _last_defaults  : ClassVar[list[str]] = []

    def __init__(self, *, first:Maybe[list[str]]=None, last:Maybe[list[str]]=None, **kwargs):
        super().__init__(**kwargs)
        self._firsts   = first or self._first_defaults
        self._lasts    = last or self._last_defaults
        self._stem_re  = re.compile("^[a-zA-Z_]+")

    def on_write(self):
        Never()

    def field_sort_key(self, field:Field) -> str|tuple:
        match self._stem_re.match(field.key):
            case None:
                key = field.key
            case x:
                key = x[0]
        try:
            return (self._lasts.index(key), field.key)
        except ValueError:
            return key

    def transform_Entry(self, entry:Entry, library:Library) -> list[Entry]:
        # Get the firsts in order if they exist
        firsts = [y for x in self._firsts if (y:=entry.get(x,None)) is not None]
        rest, lasts = [], []
        for field in entry.fields:
            match self._stem_re.match(field.key):
                case None:
                    key = field.key
                case x:
                    key = x[0]
            if key in self._firsts:
                continue
            if key not in self._lasts:
                rest.append(field)
            else:
                lasts.append(field)

        # Sort the lasts
        rest  = sorted(rest, key=self.field_sort_key)
        lasts = sorted(lasts, key=self.field_sort_key)
        entry.fields = firsts + rest + lasts
        return [entry]



#!/usr/bin/env python3
"""


"""
# ruff: noqa:
# mypy: disable-error-code="attr-defined"

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

import bibble._interface as API
from bibble.model import MetaBlock

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from bibtexparser import Library
    from bibble._interface import Middleware

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class Runner_m:
    """
    Shared code for running middlewares
    """

    def _run_writewares(self, library:Library, *, append:Maybe[list[Middleware]]=None) -> Library:
        """ Run write transforms on the library before writing,
        can handle bidirectional middlewares
        """
        append     = append or []
        fail_count = len(library.failed_blocks)
        for middleware in itz.chain(self._middlewares, append):
            fail_count = len(library.failed_blocks)
            self._logger.debug("- Running Write Middleware: %s", middleware.metadata_key())
            if hasattr(middleware, "handle_meta_entry"):
                middleware.handle_meta_entry(library)
            match middleware:
                case API.Middleware_p():
                    library = middleware.transform(library=library)
                case API.BidirectionalMiddleware_p():
                    library = middleware.write_transform(library=library)
                case x:
                    raise TypeError(type(x))

            if fail_count < (new_fcount:=len(library.failed_blocks)):
                self._logger.debug("Added %s failures", new_fcount - fail_count)
                fail_count = new_fcount

        else:
            self._record_transform_chain("write_transforms", library, append)
            return library

    def _run_readwares(self, library:Library, *, append:Maybe[list[Middleware]]=None) -> Library:
        append     = append or []
        fail_count = len(library.failed_blocks)
        for middleware in itz.chain(self._middlewares, append):
            self._logger.debug("- Running Read Middleware: %s", middleware.metadata_key())
            if hasattr(middleware, "handle_meta_entry"):
                middleware.handle_meta_entry(library)
            match middleware:
                case API.Middleware_p():
                    library = middleware.transform(library=library)
                case API.BidirectionalMiddleware_p():
                    library = middleware.read_transform(library=library)
                case x:
                    raise TypeError(type(x))

            if fail_count < (new_fcount:=len(library.failed_blocks)):
                self._logger.debug("Added %s failures", new_fcount - fail_count)
                fail_count = new_fcount

        else:
            self._record_transform_chain("read_transforms", library, append)
            return library

    def _record_transform_chain(self, meta_key:str, library:Library, append:list[Middleware]) -> None:
        """
        Record the metadata keys used on this library in a meta block
        """
        keys = [x.metadata_key() for x in itz.chain(self._middlewares, append)]
        match MetaBlock.find_in(library):
            case None:
                library.add(MetaBlock(read_stack=keys))
            case MetaBlock() as mb if meta_key in mb.data:
                mb.data[meta_key] += keys
            case MetaBlock() as mb:
                mb.data[meta_key] = keys
            case x:
                raise TypeError(type(x))




#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser import exceptions as bexp
from bibtexparser import middlewares as ms
from bibtexparser.model import MiddlewareErrorBlock
from pylatexenc.latexencode import (RULE_REGEX, UnicodeToLatexConversionRule,
                                    UnicodeToLatexEncoder)

from jgdv import Proto, Mixin
# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from . import _interface as LAPI
from ._util import UnicodeHelper_m
from bibble.util.mixins import ErrorRaiser_m, FieldMatcher_m
from bibble.util.str_transform_m import StringTransform_m
from bibble.util.middlecore import IdenBlockMiddleware

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Result
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type U2LRule = UnicodeToLatexConversionRule
    type Entry = model.Entry
    type Field = model.Field
    type Block = model.Block
    from bibtexparser.library import Library
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--|

##--|

@Proto(API.WriteTime_p, API.StrTransformer_p)
@Mixin(UnicodeHelper_m, FieldMatcher_m, StringTransform_m)
class LatexWriter(IdenBlockMiddleware):
    """ Unicode->Latex Transform.
    all strings in the library except urls, files, dois and crossrefs
    see https://pylatexenc.readthedocs.io/en/latest/latexencode/

    to customize the conversion rules, use pylatexenc.latexencode
    and call rebuild_encoder with them

    """

    _blacklist = ("url", "file", "doi", "crossref")

    def __init__(self, **kwargs):
        kwargs.setdefault(API.ALLOW_INPLACE_MOD_K, False)
        super().__init__(**kwargs)
        self.set_field_matchers(black=self._blacklist, white=[])
        self._total_options               = {}
        self._total_rules : list[U2LRule] = [
            self.build_encode_rule(LAPI.ENCODING_RULES)
        ]
        if kwargs.get(API.KEEP_MATH_K, True):
            self._total_rules.append(self.build_encode_rule(LAPI.MATH_RULES))

        if kwargs.get(API.ENCLOSE_URLS_K, False):
            self._total_rules.append(self.build_encode_rule(LAPI.URL_RULES))

        self.rebuild_encoder()

    def on_write(self):
        Never()

    def transform_Entry(self, entry: Entry, library: Library) -> list[Block]:
        match self.match_on_fields(entry, library):
            case model.Entry() as x:
                return [x]
            case Exception() as err:
                return [self.make_error_block(entry, err)]
            case x:
                raise TypeError(type(x))

    def field_h(self, field:model.Field, entry) -> Result[list[Field], Exception]:
        match self.transform_strlike(field.value):
            case Exception() as err:
                return err
            case x:
                return [model.Field(key=field.key, value=x)]

    def _transform_raw_str(self, python_string: str) -> Result[str, Exception]:
        try:
            return self._encoder.unicode_to_latex(python_string)
        except Exception as e:
            return e



#!/usr/bin/env python3

__version__ = "0.1.1"

from . import model
from bibble.library import BibbleLib
from bibble.util.pair_stack import PairStack
from bibble import metadata as metadata
from bibble import bidi
from bibble import files
from bibble import fields
from bibble import latex
from bibble import people
from bibble import failure
from bibble import util
from bibble import io



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

from bibtexparser import model
from bibtexparser.library import Library
from bibble.io.writer import BibbleWriter

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Block = model.Block
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class RstWriter(BibbleWriter):
    """ Write bibtex entries as Rst. """

    # These need to match the BibEntryDirective of the sphinx domain
    _entry      : ClassVar[str]       = ".. bibtex:entry:: {}"
    _entry_args : ClassVar[list[str]] = ["title","author", "editor", "year", "tags",
                                         "journal", "booktitle", "within",
                                         "platform", "publisher", "institution",
                                         "series", "url", "doi", "isbn", "edition",
                                         "crossref"
                                         ]
    _indent     : ClassVar[str]       = " "*3

    def _title_add(self, block:Block) -> list[str]:
        """ Format and return the title """
        match block.get('title', None), block.get("subtitle", None):
            case model.Field(value=str() as title), model.Field(value=str() as subtitle):
                return [f"{self._indent}:title: {title}: {subtitle}\n"] # type:ignore
            case model.Field(value=str() as title), _:
                return [f"{self._indent}:title: {title}\n"] # type: ignore
                pass
            case _:
                raise KeyError("no title", block.key)

    def _must_add(self, block:Block, field:str) -> list[str]:
        match block.get(field, None):
            case model.Field(value=val):
                return [f"{self._indent}:{block.key}: {val}\n"] # type: ignore
            case _:
                raise KeyError('Entry missing required field', block.key, field)

    def _can_add(self, block:Block, *keys:str) -> list[str]:
        result = []
        for key in keys:
            match block.get(key, None):
                case None:
                    continue
                case model.Field(value=val):
                    result.append(f"{self._indent}:{key}: {val}\n") # type: ignore

        else:
            return result

    def visit_entry(self, block:Block) -> list[str]:
        result = [self._entry.format(block.key)]
        match block.entry_type:
            case "case" | "legal" | "judicial" | "law":
                return []
            case "standard" | "online" | "blog" | "dataset":
                return []
            case "tweet" | "thread":
                return []
            case _:
                result += self._title_add(block)
                result += self._must_add(block, "tags")
                result += self._can_add(block, "author", "editor", "year", "series")
                result += self._can_add(block, "journal", "booktitle", "doi", "url", "isbn", "publisher")
                result += self._can_add(block, "incollection", "institution")
                # TODO volume, number, pages, chapter

        result += ["\n\n..\n",
                   f"{self._indent} Fields:\n",
                   "{} {}\n".format(self._indent, ", ".join(block.fields_dict.keys())),
                   f"{self._indent} Object Keys:\n",
                   "{} {}\n".format(self._indent,
                                    ", ".join([x for x in dir(block) if "__" not in x])),
                   "\n\n",
                    ]

        return result

    def make_header(self, library:Library, file:Maybe[pl.Path]=None) -> list[str]:
        match file:
            case None:
                title = "A Bibtex File"
            case pl.Path():
                title = file.stem

        lines = [".. -*- mode: ReST -*-\n\n",
                 f".. _{title}:\n\n",
                 "="*len(title), "\n",
                 title, "\n",
                 "="*len(title), "\n\n",
                 ".. contents:: Entries:\n",
                 "   :class: bib_entries\n",
                 "   :local:\n\n",
                 "For the raw bibtex, see `the file`_.\n\n",
                 f".. _`the file`: https://github.com/jgrey4296/bibliography/blob/main/main/{title}.bib\n\n",
                 ]
        return lines



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware, LibraryMiddleware)
from jgdv import Proto, Mixin

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from bibble.util.mixins import ErrorRaiser_m, FieldMatcher_m
from bibble.util.middlecore import IdenBlockMiddleware

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(API.ReadTime_p)
@Mixin(ErrorRaiser_m, FieldMatcher_m)
class PathReader(IdenBlockMiddleware):
    """
      Convert file paths in bibliography to pl.Path's, expanding relative paths
      according to lib_root
    """

    _whitelist = ("file",)

    def __init__(self, *, lib_root:Maybe[pl.Path]=None, **kwargs):
        super().__init__(**kwargs)
        self._lib_root = lib_root or pl.Path.cwd()
        self.set_field_matchers(white=kwargs.pop("whitelist", self._whitelist), black=[])  # type: ignore[attr-defined]

    def on_read(self):
        Never()

    def transform_Entry(self, entry, library):
        match self.match_on_fields(entry, library):
            case model.Entry() as x:
                return [x]
            case Exception() as err:
                return [self.make_error_block(entry, err)]
            case x:
                raise TypeError(type(x))

    def field_h(self, field, entry):
        base = pl.Path(field.value)
        match base.parts[0]:
            case "/":
                field.value = base
            case "~":
                field.value = base.expanduser().resolve()
            case _:
                field.value = self._lib_root / base

        if not field.value.exists():
            return ValueError(f"File does not exist: {field.value}")

        return [field]



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import atexit#  for @atexit.register
import collections
import contextlib
import datetime
import enum
import faulthandler
import functools as ftz
import hashlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser import exceptions as bexp
from bibtexparser import middlewares as ms
from bibtexparser.model import MiddlewareErrorBlock

# ##-- end 3rd party imports

# ##-- 1st party imports
from bibble import _interface as API
import bibble.model as bmodel

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Either, Result, Rx
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
EMPTY     : Final[Rx]  = re.compile(r"^$")
OR_REG    : Final[str] = r"|"
ANY_REG   : Final[Rx]  = re.compile(r".")
# Body:

class MiddlewareValidator_m:
    """ For ensuring the middlewares of a reader/writer  are appropriate,
    by excluding certain middlewares.
    """

    def exclude_middlewares(self, proto:Protocol):
        if not issubclass(proto, Protocol):
            raise TypeError("Tried to validate middlewares with a non-protocol", proto)
        failures = []
        for middle in self._middlewares:
            # note: no 'not'.
            if isinstance(middle, proto):
                failures.append(middle)
            elif not isinstance(middle, API.Middleware_p|API.BidirectionalMiddleware_p):
                failures.append(middle)
        else:
            if bool(failures):
                raise TypeError("Bad middlewares", failures)

class ErrorRaiser_m:
    """ Mixin for easily combining middleware errors into a block"""

    def make_error_block(self, entry:API.Entry, err:Exception) -> MiddlewareErrorBlock:
        return bmodel.FailedBlock(block=entry, error=err, source=self)

class FieldMatcher_m:
    """ Mixin to process fields if their key matchs a regex
    defaults are in the attrs _field_blacklist and _field_whitelist, _entry_whitelist
    Call set_field_matchers to extend.
    Call match_on_fields to start.
    Call maybe_skip_entry to compare the lowercase entry type to a whitelist
    Implement field_handler to use.

    match_on_fields calls entry.set_field on the field_handlers result
    """

    def set_field_matchers(self, *, white:list[str], black:list[str]) -> Self:
        """ sets the blacklist and whitelist regex's
        returns self to help in building parse stacks
        """
        match white:
            case []:
                self._field_white_re = ANY_REG
            case [*xs]:
                self._field_white_re = re.compile(OR_REG.join(xs))

        match black:
            case []:
                self._field_black_re = EMPTY
            case [*xs]:
                self._field_black_re = re.compile(OR_REG.join(xs))

        return self

    def match_on_fields(self, entry: API.Entry, library: API.Library) -> Result[API.Entry, Exception]:
        errors = []
        whitelist, blacklist = self._field_white_re, self._field_black_re
        for field in entry.fields:
            match field:
                case model.Field(key=key) if whitelist.match(key) and not blacklist.match(key):
                    res = self.field_h(field, entry)
                case _:
                    continue

            match res:
                case [*xs]:
                    for x in xs:
                        entry.set_field(x)
                case Exception() as err:
                    errors += err.args
                case x:
                    raise TypeError(type(x), self)
        else:
            if bool(errors):
                return ValueError(*errors)

            return entry

    def field_h(self, field:API.Field, entry:API.Entry) -> Result[list[API.Field], Exception]:
        raise NotImplementedError("Implement the field handler")

class EntrySkipper_m:
    """
    Be able to skip entries by their type
    """

    def set_entry_skiplists(self, *, white:list[str]=None, black:list[str]=None) -> None:
        self._entry_whitelist = [x.lower() for x in white or []]
        self._entry_blacklist = [x.lower() for x in black or []]

    def should_skip_entry(self, entry:API.Entry, library:API.Library) -> bool:
        low_type = entry.entry_type.lower()
        match self._entry_blacklist:
            case list() as xs if low_type in xs:
                return True
            case _:
                pass

        match self._entry_whitelist:
            case []:
                return False
            case list() as xs if low_type in xs:
                return False
            case _:
                return True



#!/usr/bin/env python3
"""
Refactored core middleware classes from bibtexparser

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
import sys
# ##-- end stdlib imports

import tqdm
import bibble._interface as API
from bibble.model import MetaBlock
import jgdv
from jgdv import Proto
from bibtexparser.library import Library
from bibtexparser import model

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Logger          = logmod.Logger
    type Block           = model.Block
    type Entry           = model.Entry
    type String          = model.String
    type Preamble        = model.Preamble
    type ExplicitComment = model.ExplicitComment
    type ImplicitComment = model.ImplicitComment
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

@Proto(jgdv.protos.DILogger_p)
class _BaseMiddleware:
    """
    The base middleware.
    has a metadata key,
    stores allow_inplace and allow_parallel
    and can have an injected logger.

    Any extra init kwargs are stored in _extra
    """
    allow_inplace : bool
    allow_parallel : bool

    @classmethod
    def metadata_key(cls) -> str:
        """Identifier of the middleware.
        This key is used to identify the middleware in a blocks metadata.
        """
        return f"bibble-{cls.__name__}"

    def __init__(self, **kwargs):
        """

        """
        inplace             = kwargs.pop(API.ALLOW_INPLACE_MOD_K, True)
        self.allow_inplace  = inplace
        self.allow_parallel = kwargs.pop(API.ALLOW_PARALLEL_K, False)
        fallback_name       = f"{type(self).__module__}.{type(self).__name__}"
        self._logger        = kwargs.pop(API.LOGGER_K, logmod.getLogger(fallback_name))
        self._extra         = kwargs

    def logger(self) -> Logger:
        return self._logger

    def _get_lib_iterator(self, library:Library) -> tuple[Library, Iterator[Block]]:
        match self.allow_inplace:
            case True:
                library = library
            case False:
                library = deepcopy(library)
            case x:
                raise TypeError(type(x))

        match self._extra:
            case {"tqdm":True} if sys.stdout.isatty():
                iterator = tqdm.tqdm(enumerate(library.blocks),
                                     desc=type(self).__name__,
                                     total=len(library.blocks),
                                     ncols=API.TQDM_WIDTH)
            case _:
                iterator = enumerate(library.blocks)

        return library, iterator

    def handle_meta_entry(self, library:Library) -> None:
        """ An optional entry hook for middlewares,
        which is given the library's metablock before transform is called.

        Use case: conditionally setting PathWriter suppress paths
        """
        pass
class IdenLibraryMiddleware(_BaseMiddleware):
    """ Identity Library Middleware, does nothing """

    def transform(self, library:Library) -> Library:
        match library:
            case Library() if self.allow_inplace:
                return library
            case Library():
                return deepcopy(library)

@Proto(API.AdaptiveMiddleware_p, API.Middleware_p)
class IdenBlockMiddleware(_BaseMiddleware):
    """ Identity Block Middleware, does nothing
    If passed 'tqdm'=True uses tqdm around the block level loop
    """
    _transform_cache : dict[type, list[Callable]]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._transform_cache = dict()

    def get_transforms_for(self, block:Block, *, direction:Maybe[str]=None) -> list[Callable]:
        """ Get all transforms of the form transform_{Type},
        by mro, from most -> least specific
        """
        assert(direction is None)
        if type(block) not in self._transform_cache:
            mro = type(block).mro()
            formatted = [f"transform_{x.__name__}" for x in mro]
            methods = [getattr(self, x, None) for x in formatted]
            self._transform_cache[type(block)] = [x for x in methods if x is not None]

        return self._transform_cache[type(block)]

    def transform(self, library:Library) -> Library:
        library, iterator = self._get_lib_iterator(library)
        blocks = []
        for i,block in iterator:
            match self.get_transforms_for(block):
                case [x, *_]: # Use the first found transform,
                    transform = x
                case _: # No transforms for this block type, do nothing
                    blocks.append(block)
                    continue

            match transform(block, library):
                case None: # remove block
                    continue
                case []: # Keep original (it might have been modified)
                    blocks.append(block)
                case [*xs]: # new blocks
                    blocks += xs
                case x:
                    raise TypeError(type(x), i, block)

        else:
            # Remove the old blocks
            library.remove(library.blocks[:])
            # Add the new blocks
            library.add(blocks)
            return library

@Proto(API.AdaptiveMiddleware_p, API.BidirectionalMiddleware_p)
class IdenBidiMiddleware(_BaseMiddleware):

    _transform_cache : dict[str, list[Callable]]
    _reader : Maybe[API.Middleware]
    _writer : Maybe[API.Middleware]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._reader          = None
        self._writer          = None
        self._transform_cache = dict()


    def handle_meta_entry(self, library:Library) -> None:
        match self._reader:
            case None:
                pass
            case mw if not hasattr(mw, "handle_meta_entry"):
                pass
            case mw:
                mw.handle_meta_entry(library)

        match self._writer:
            case None:
                pass
            case mw if not hasattr(mw, "handle_meta_entry"):
                pass
            case mw:
                mw.handle_meta_entry(library)


    def get_transforms_for(self, block:Block, *, direction:Maybe[str]=None) -> list[Callable]:
        """ Get all transforms of the form {direction}_transform_{Type},
        by mro, from most -> least specific
        """
        cache_key = f"{direction}_{type(block).__name__}"
        if cache_key not in self._transform_cache:
            mro       = type(block).mro()
            formatted = [f"{direction}_transform_{x.__name__}" for x in mro]
            methods   = [getattr(self, x, None) for x in formatted]
            self._transform_cache[cache_key] = [x for x in methods if x is not None]

        return self._transform_cache[cache_key]


    def read_transform(self, library:Library) -> Library:
        library, iterator = self._get_lib_iterator(library)
        blocks   = []
        for i,block in iterator:
            match self.get_transforms_for(block, direction="read"):
                case [x, *_]: # Use the first found transform,
                    transform = x
                case _:
                    blocks.append(block)
                    continue

            match transform(block, library):
                case [] | None: # Transform gave nothing, so keep the original block.
                    blocks.append(block)
                case [*xs]:
                    # new blocks
                    blocks += xs
                case x:
                    raise TypeError(type(x), i, block)
        else:
            # Remove the old blocks
            library.remove(library.blocks[:])
            assert(len(library.blocks) == 0), len(library.blocks)
            # Add the new blocks
            library.add(blocks)
            return library

    def write_transform(self, library:Library) -> Library:
        library, iterator = self._get_lib_iterator(library)
        blocks = []
        for i,block in iterator:
            match self.get_transforms_for(block, direction="write"):
                case [x, *_]: # Use the first found transform,
                    transform = x
                case _:
                    blocks.append(block)
                    continue

            match transform(block, library):
                case [] | None: # Transform gave nothing, so keep the original block.
                    blocks.append(block)
                case [*xs]: # new blocks
                    blocks += xs
                case x:
                    raise TypeError(type(x), i, block)
        else:
            # Remove the old blocks
            library.remove(library.blocks[:])
            # Add the new blocks
            library.add(blocks)
            return library



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser.middlewares.names import NameParts
from jgdv import Mixin, Proto

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from bibble.util.middlecore import IdenBlockMiddleware
from bibble.util.mixins import FieldMatcher_m
from bibble.util.name_parts import NameParts_d

# ##-- end 1st party imports

from . import _interface as API_N

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Result
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Entry = model.Entry
    type Field = model.Field
    from bibtexparser import Library

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(API.WriteTime_p)
@Mixin(FieldMatcher_m)
class NameWriter(IdenBlockMiddleware):
    """ Transforms NameParts -> [str] -> str

    on init:
    - parts : bool    = merge name parts together
    - authors : bool  = merge separate authors together
    """
    _whitelist = ("author", "editor", "translator")

    def __init__(self, *, parts:bool, authors:bool, **kwargs):
        super().__init__(**kwargs)
        self._merge_parts   = parts
        self._merge_authors = authors
        self.set_field_matchers(white=self._whitelist, black=[])

    def on_write(self):
        Never()

    def transform_Entry(self, entry:Entry, library:Library) -> list[Entry]:
        match self.match_on_fields(entry, library):
            case model.Entry() as x:
                return [x]
            case Exception() as err:
                return [self.make_error_block(entry, err)]
            case x:
                raise TypeError(type(x))

    def field_h(self, field:Field, entry:Entry) -> Result[list[Field], Exception]:
        result = []
        merger = self._merge_von_last_jr_first
        match field.value, self._merge_parts, self._merge_authors:
            case str(), _, _:
                pass
            case [*xs], True, False if all(isinstance(x, NameParts|NameParts_d) for x in xs):
                # Merge nameparts, but not authors
                merged = [merger(x) for x in xs]
                result.append(model.Field(field.key, merged, start_line=field.start_line))
            case [*xs], True, True if all(isinstance(x, NameParts|NameParts_d) for x in xs):
                # Merge nameparts, then authors
                merged = [merger(x) for x in xs]
                joined = API_N.JOIN_STR.join(merged)
                result.append(model.Field(field.key, joined, start_line=field.start_line))
            case [*xs], _, True if all(isinstance(x, str) for x in xs) and self._merge_authors:
                # Just merge authors
                joined = API_N.JOIN_STR.join(xs)
                result.append(model.Field(field.key, joined, start_line=field.start_line))
            case _, False, False:
                pass
            case x, _, _:
                return TypeError("Unexpected name type", entry.key, type(x))

        return result

    def _merge_von_last_jr_first(self, name:NameParts|NameParts_d) -> str:
        match name:
            case NameParts() | NameParts_d():
                pass
            case x:
                raise TypeError(type(x))

        result = []
        if name.von:
            result.append(" ".join(name.von))
            result.append(" ")

        if name.last:
            result.append(" ".join(name.last))
            result.append(", ")

        if name.jr:
            result.append(" ".join(name.jr))
            result.append(", ")

        result.append(" ".join(name.first))

        full_name = "".join(result).removesuffix(", ").strip()
        return full_name



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)
from jgdv import Proto
# ##-- end 3rd party imports

import bibble._interface as API
from bibble.util.middlecore import IdenLibraryMiddleware
import bibble.model as bmodel

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Logger = logmod.Logger
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class FailureHandler(IdenLibraryMiddleware):
    """ Middleware to Filter failed blocks of a library,
    either to a logger output, or to a file
    Put at end of parse stack

    Will log out where the failed blocks start by line.
    """

    def __init__(self, **kwargs):
        file_target = kwargs.pop("file", None)
        super().__init__(**kwargs)
        match file_target:
            case str() as x:
                self.file_target = pl.Path(x)
            case pl.Path() as x:
                self.file_target = x
            case _:
                self.file_target = None


    def transform(self, library):
        total    = len(library.failed_blocks)
        reported = []
        for i, block in enumerate(library.failed_blocks, start=1):
            source_file = self._find_source_file(block, library)
            match block:
                case bmodel.FailedBlock():
                    report = block.report(i=i, total=total, source_file=source_file)[0]
                case model.ParsingFailedBlock() if source_file:
                    report = f"({i}/{total}) Bad Block: : {block.start_line} : {source_file} : {block.error}"
                case model.ParsingFailedBlock():
                    report = f"({i}/{total}) Bad Block: : {block.start_line} : {block.error}"
                case x:
                    raise TypeError(type(x))
            ##--|
            reported.append((report, block.error, block.raw))
            self._logger.info(report)
        else:
            self.write_failures_to_file(reported)
        return library

    def write_failures_to_file(self, reports:list) -> None:
        match reports:
            case []:
                return
            case _ if self.file_target is None:
                return
            case _:
                pass

        total = len(reports)
        with self.file_target.open("w") as f:
            for rep, err, raw in reports:
                f.writelines(["\n\n--------------------\n",
                               rep,
                               f"\nError: {err}",
                               "\n--------------------\n",
                               raw,
                               ])

    def _find_source_file(self, block, library) -> Maybe[str]:
        match block:
            case model.Entry():
                target_key = block.key
            case _:
                return None

        match bmodel.MetaBlock.find_in(library):
            case None:
                return None
            case bmodel.MetaBlock() as meta if 'sources' not in meta.data:
                return None
            case bmodel.MetaBlock() as meta:
                data = meta.data
                sources = meta.data['sources']

        for source in sources:
            if source in data and target_key in data[source]:
                return source
            ##--|
        else:
            return None



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from jgdv import Mixin, Proto
from jgdv.debugging.timeblock_ctx import TimeBlock_ctx
from bibtexparser.library import Library
from bibtexparser.splitter import Splitter

# ##-- end 3rd party imports

from bibble import _interface as API
from bibble.model import MetaBlock
from bibble.util.mixins import MiddlewareValidator_m
from bibble.util import PairStack
from ._util import Runner_m

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Logger = logmod.Logger
    type Middleware = API.Middleware_p | API.BidirectionalMiddleware_p
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(API.Reader_p)
@Mixin(Runner_m, MiddlewareValidator_m)
class BibbleReader:
    """ A Refactored bibtexparser reader

    """
    _middlewares : list[Middleware]
    _lib_class   : type[Library]

    def __init__(self, stack:PairStack|list[Middleware], *, lib_base:Maybe[type]=None, logger:Maybe[Logger]=None):
        match stack:
            case PairStack():
                self._middlewares = stack.read_stack()
            case list():
                self._middlewares = stack
            case x:
                raise TypeError(type(x))

        self._lib_class : type = lib_base or Library
        self._logger           = logger or logging

        self.exclude_middlewares(API.WriteTime_p)
        if not issubclass(self._lib_class, Library):
            raise TypeError("Bad library base pased to reader", lib_base)

    def read_dir(self, source:pl.Path, *, ext:str, into:Maybe[Library]=None, append:Maybe[list[Middleware]]=None) -> Maybe[Library]:
        visited : set = set()
        to_read : list = []
        for args in pl.Path().walk(top_down=True, on_error=None, follow_symlinks=False):
            dpath     : pl.Path   = args[0]
            dnames    : list[str] = args[0] # Edit to control descent
            filenames : list[str] = args[2]
            if dpath in visited:
                dnames.clear()
            else:
                visited.add(dpath)

            to_read += [y for x in filenames if (y:=dpath/x).suffix == ext]
        else:
            pass
        ##--|
        lib = into or self._lib_class()
        for x in to_read:
            match self.read(x, into=lib, append=append):
                case None:
                    return None
                case Library() as y:
                    lib = y
        else:
            return lib

    def read(self, source:str|pl.Path, *, into:Maybe[Library]=None, append:Maybe[list[Middleware]]=None) -> Maybe[Library]:
        """ read source and make a new library.
        if given 'into' lib, add the newly read entries into that libray as well
        """
        source_text : str
        basic       : Library
        transformed : Library

        match source:
            case str():
                source_text = source
            case pl.Path():
                try:
                    source_text = source.read_text()
                except UnicodeDecodeError as err:
                    logging.exception("Unicode Error in File: %s, Start: %s", source, err.start)
                    return None
            case x:
                raise TypeError(type(x))

        with TimeBlock_ctx(enter="--> Bibtex Reading: Start",
                           exit="<-- Bibtex Reading:",
                           level=logmod.INFO):
            basic       = self._read_into(self._lib_class(), source_text)

        with TimeBlock_ctx(enter="--> Read Transforms: Start",
                           exit="<-- Read Transforms:",
                           level=logmod.INFO):
            transformed = self._run_readwares(basic, append=append)

        entry_keys : set = {x.key for x in transformed.entries}
        match into:
            case Library():
                 into.add(transformed.blocks)
                 final_lib = into
            case None:
                final_lib = transformed
            case x:
                raise TypeError(type(x))

        return self._map_keys(final_lib, source, entry_keys)

    def _map_keys(self, final_lib:Library, source:str|pl.Path, entry_keys:set[str]) -> Library:
        """ Map source -> keys

        """
        logging.debug("Mapping %s new keys to source %s", len(entry_keys), source)
        match MetaBlock.find_in(final_lib), source:
            case None, str():
                final_lib.add(MetaBlock(sources={"raw_text"}, raw_text=entry_keys))
            case None, pl.Path():
                kwargs = {str(source) : entry_keys}
                final_lib.add(MetaBlock(sources={source}, **kwargs))
            case MetaBlock() as b, str() if 'sources' in b.data:
                b.data['sources'].add("raw_text")
                b.data["raw_text"] = entry_keys
            case MetaBlock() as b, pl.Path() if 'sources' in b.data:
                b.data['sources'].add(source)
                b.data[str(source)] = entry_keys
            case MetaBlock() as b, str():
                b.data['sources']  = {"raw_text"}
                b.data["raw_text"] = entry_keys
            case MetaBlock() as b, pl.Path():
                b.data['sources'] = {source}
                b.data[source] = entry_keys
            case x:
                raise TypeError(type(x))

        return final_lib

    def _read_into(self, lib:Library, source:str) -> Library:
        assert(isinstance(source, str))
        splitter = Splitter(bibstr=source)
        library  = splitter.split(library=lib)
        return library




#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

from bibtexparser.library import Library
from bibtexparser import model
import bibble._interface as API
from .. import _interface as API_F
from bibble.fields import TitleCleaner, TitleSplitter
# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class TestTitleCleaner:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match TitleCleaner():
            case TitleCleaner():
                assert(True)
            case x:
                 assert(False), x

    def test_clean(self):
        initial_title =  "   a test title with whitespace   "
        mid    = TitleCleaner()
        entry1 = model.Entry("test", "first",  [
            model.Field("title", initial_title),
            model.Field("subtitle", initial_title),
        ])
        lib    = Library()
        lib.add(entry1)
        assert(mid.transform(lib) is lib)
        assert(initial_title.endswith(" "))
        assert(entry1.fields_dict['title'].value == initial_title.strip())
        assert(entry1.fields_dict['subtitle'].value == initial_title.strip())

    def test_clean_ignores_other_fields(self):
        initial_title =  "  a test title with whitespace   "
        mid    = TitleCleaner()
        entry1 = model.Entry("test", "first",  [
            model.Field("title", initial_title),
            model.Field("not_title", initial_title),
        ])
        lib    = Library()
        lib.add(entry1)
        assert(mid.transform(lib) is lib)
        assert(initial_title.endswith(" "))
        assert(entry1.fields_dict['title'].value == initial_title.strip())
        assert(entry1.fields_dict['not_title'].value == initial_title)

class TestTitleSplitter:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match TitleSplitter():
            case API.AdaptiveMiddleware_p():
                assert(True)
            case x:
                 assert(False), x

    def test_clean(self):
        initial_title =  "  a test title with whitespace   "
        mid    = TitleSplitter()
        entry1 = model.Entry("test", "first",  [
            model.Field("title", initial_title),
            model.Field("subtitle", initial_title),
        ])
        lib    = Library()
        lib.add(entry1)
        assert(mid.transform(lib) is lib)
        assert(initial_title.endswith(" "))
        assert(entry1.fields_dict['title'].value == initial_title.strip())
        assert(entry1.fields_dict['subtitle'].value == initial_title.strip())

    def test_clean_ignores_other_fields(self):
        initial_title =  "  a test title with whitespace   "
        mid    = TitleSplitter()
        entry1 = model.Entry("test", "first",  [
            model.Field("title", initial_title),
            model.Field("not_title", initial_title),
        ])
        lib    = Library()
        lib.add(entry1)
        assert(mid.transform(lib) is lib)
        assert(initial_title.endswith(" "))
        assert(entry1.fields_dict['title'].value == initial_title.strip())
        assert(entry1.fields_dict['not_title'].value == initial_title)

    def test_split(self):
        initial_title =  "  a test title with whitespace : the subtitle   "
        target = "a test title with whitespace"
        target_sub = "the subtitle"
        mid    = TitleSplitter()
        entry1 = model.Entry("test", "first",  [
            model.Field("title", initial_title),
        ])
        lib    = Library()
        lib.add(entry1)
        assert(mid.transform(lib) is lib)
        assert(initial_title.endswith(" "))
        assert(entry1.fields_dict['title'].value == target)
        assert(entry1.fields_dict['subtitle'].value == target_sub)

    def test_split_doesnt_override_subtitle(self):
        initial_title =  "  a test title with whitespace : the subtitle   "
        initial_sub   =  " diff subtitle "
        target_sub    = "diff subtitle"
        mid           = TitleSplitter()
        entry1        = model.Entry("test", "first",  [
            model.Field("title", initial_title),
            model.Field("subtitle", initial_sub),
        ])
        lib    = Library()
        lib.add(entry1)
        assert(mid.transform(lib) is lib)
        assert(initial_title.endswith(" "))
        assert(entry1.fields_dict['title'].value == initial_title.strip())
        assert(entry1.fields_dict['subtitle'].value == target_sub)



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from jgdv import Proto
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)

# ##-- end 3rd party imports

from bibble._interface import ReadTime_p
from bibble.util.middlecore import IdenBlockMiddleware
from . import _interface as API_F

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Entry = model.Entry
    from bibtexparser.library import Library

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--|

@Proto(ReadTime_p)
class TitleCleaner(IdenBlockMiddleware):
    """
      strip whitespace from the title, and (optional) subtitle
    """

    def on_read(self):
        Never()

    def transform_Entry(self, entry, library):
        match entry.get(API_F.TITLE_K):
            case None:
                self._logger.warning("Entry has no title: %s", entry.key)
            case model.Field(value=str() as value):
                entry.set_field(model.Field(API_F.TITLE_K, value.strip()))
            case _:
                pass

        match entry.get(API_F.SUBTITLE_K):
            case None:
                pass
            case model.Field(value=str() as value):
                entry.set_field(model.Field(API_F.SUBTITLE_K, value.strip()))
            case _:
                pass

        return [entry]

@Proto(ReadTime_p)
class TitleSplitter(IdenBlockMiddleware):
    """
      Split Title Into Title and Subtitle, If Subtitle Doesn't Exist Yet

    strips whitespace as well
    """

    def on_read(self):
        Never()

    def transform_Entry(self, entry:Entry, library:Library):
        match entry.get(API_F.TITLE_K), entry.get(API_F.SUBTITLE_K):
            case None, _:
                self._logging.warning("Entry has no title: %s", entry.key)
            case model.Field(value=title), model.Field(value=subtitle):
                entry.set_field(model.Field(API_F.TITLE_K, title.strip()))
                entry.set_field(model.Field(API_F.SUBTITLE_K, subtitle.strip()))
                pass
            case model.Field(value=value), None if API_F.TITLE_SEP in value:
                title, *rest = value.split(API_F.TITLE_SEP)
                entry.set_field(model.Field(API_F.TITLE_K, title.strip()))
                entry.set_field(model.Field(API_F.SUBTITLE_K, " ".join(rest).strip()))
            case model.Field(value=value), None:
                entry.set_field(model.Field(API_F.TITLE_K, value.strip()))
            case None, None:
                pass

        return [entry]



"""

"""

from .title_reader import TitleCleaner, TitleSplitter
from .url_reader import CleanUrls, ExpandUrls
from .field_substitutor import FieldSubstitutor
from .field_sorter import FieldSorter
from .field_accumulator import FieldAccumulator



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from collections import defaultdict
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from jgdv import Proto
from bibtexparser.library import Library
from bibtexparser.middlewares import BlockMiddleware

# ##-- end 3rd party imports

import bibble._interface as API

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from API import Middleware
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(API.Library_p)
class BibbleLib(Library):
    """ A library with a key value store for extra info
    Also tracks the individual files used as source
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._kv_store    = defaultdict(set)
        self.source_files = set()

    def add_sublibrary(self, lib:Library, source:Maybe[str|pl.Path]=None) -> Self:
        """ Merge entries, kv_store and source files into this library
        will *overwrite* existing kv_store keys
        """
        if source:
            self.source_files.add(source)

        match lib:
            case BibbleLib():
                self.add(lib.entries)
                self.source_files.update(lib.source_files)
                for k,v in lib._kv_store.items():
                    self.store_meta_value(k, v)
            case Library():
                self.add(lib.entries)
            case _:
                raise TypeError("Bad update sublibrary")

        return self

    def store_meta_value(self, key:str|Middleware, value:Any):
        raise DeprecationWarning("Use a MetaBlock")

    def get_meta_value(self, key) -> set|Any:
        raise DeprecationWarning("Use a MetaBlock")



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from bibtexparser import model
from bibtexparser.library import Library

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Either, Result
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from bibble.util import NameParts_d
    from bibtexparser.middlewares import NameParts
    type StringBlock     = model.String
    type Preamble        = model.Preamble
    type Field           = model.Field
    type Block           = model.Block
    type Entry           = model.Entry
    type String          = model.String
    type FailedBlock     = model.ParsingFailedBlock
    type ErrorBlock      = model.MiddlewareErrorBlock
    type CommentBlock    = model.ExplicitComment | model.ImplicitComment
    type UniMiddleware   = Middleware_p | AdaptiveMiddleware_p
    type BidiMiddleware  = BidirectionalMiddleware_p
    type Middleware      = UniMiddleware | BidiMiddleware

    type StrLike         = list|set|String|NameParts_d|NameParts
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
ALLOW_INPLACE_MOD_K : Final[str] = "allow_inplace_modification"
ALLOW_PARALLEL_K    : Final[str] = "allow_parallel_execution"
LOGGER_K            : Final[str] = "logger"
KEEP_MATH_K         : Final[str] = "keep_math"
ENCLOSE_URLS_K      : Final[str] = "enclose_urls"

TQDM_WIDTH          : Final[int] = 150
##--|
## Enums / Flags

class Capability_f(enum.Flag):
    """ A Flag for where middlewares can be in the read/write stack """

    insist_front = enum.auto()
    insist_end   = enum.auto()
    read_time    = enum.auto()
    write_time   = enum.auto()
    validate     = enum.auto()
    transform    = enum.auto()
    report       = enum.auto()
    dependent    = enum.auto()

##--|
## Bibtexparser protcols

@runtime_checkable
class Library_p(Protocol):
    """ The core methods of a library """

    def add(self, blocks:list[Block], fail_on_duplicate_key: bool = False) -> None:
        pass

    def remove(self, blocks:list[Block]) -> None:
        pass

    def replace(self, old_block:Block, new_block:Block, fail_on_duplicate_key:bool=True) -> None:
        pass

    def blocks(self) -> list[Block]:
        pass

    def failed_blocks(self) -> list[FailedBlock]:
        pass

    def strings(self) -> list[StringBlock]:
        pass

    def strings_dict(self) -> dict[str, StringBlock]:
        pass

    def entries(self) -> list[Entry]:
        pass

    def entries_dict(self) -> dict[str, Entry]:
        pass

    def preambles(self) -> list[Preamble]:
        pass

    def comments(self) -> list[CommentBlock]:
        pass

@runtime_checkable
class Middleware_p(Protocol):
    """ A Middleware is something with a 'transform' method """

    def transform(self, library:Library) -> Library:
        pass

@runtime_checkable
class BlockMiddleware_p(Protocol):
    """ A (non-adaptive) block middleware has 'transform_x' methods """

    def transform(self, library:Library) -> Library:
        pass

    def transform_block(self, block:Block, library:Library) -> list[Block]:
        pass

    def transform_entry(self, entry:Entry, library:Library) -> list[Block]:
        pass

    def transform_string(self, string:model.StringBlock, library:Library) -> list[Block]:
        pass

    def transform_preamble(self, preamble:model.Preamble, library:Library) -> list[Block]:
        pass

    def transform_explicit_comment(self, comment:model.ExplicitComment, library:Library) -> list[Block]:
        pass

    def transform_implicit_comment(self, comment:model.ImplicitComment, library:Library) -> list[Block]:
        pass


##--| New Middleware protocols:

@runtime_checkable
class BidirectionalMiddleware_p(Protocol):
    """ A Single middleware that holds the logic for reading and writing,
    Intended for undoing what is done on read, prior to writing.

    eg: Latex Decoding, then Encoding
    """

    def read_transform(self, library:Library) -> Library:
        pass

    def write_transform(self, library:Library) -> Library:
        pass

@runtime_checkable
class AdaptiveMiddleware_p(Protocol):
    """ Middleware that looks up defined transforms using the type name, by mro.
    The form for method lookup is either:
    - transform_{type(block).__class__}
    - {direction}_transform_{type(block).__class__}

    An adaptive middleware doesn't need all the 'transform_X' methods a BlockMiddleware_p does.

    """

    def get_transforms_for(self, block:Block, *, direction:Maybe[str]=None) -> list[Callable[[Block, Library], list[Block]]]:
        pass

##--| IO Protocols

@runtime_checkable
class PairStack_p(Protocol):
    """ Protocol for both storing both read and write middlewares

    """

    def add(self, *args:BidiMiddleware, read:Maybe[list|Middleware]=None, write:Maybe[list|Middleware]=None) -> Self:
        pass

    def read_stack(self) -> list[Middleware]:
        pass

    def write_stack(self) -> list[Middleware]:
        pass

@runtime_checkable
class Reader_p(Protocol):
    """ Readers take source text, or a file, or a directory,
    parses the read bibtex, running middlewares on the parsed bibtex
    """

    def read(self, source:str|pl.Path, *, into:Maybe[Library]=None, append:Maybe[list[Middleware]]=None) -> Maybe[Library]:
        pass

    def read_dir(self, source:pl.Path, *, ext:str, into:Maybe[Library]=None, append:Maybe[list[Middleware]]=None) -> Maybe[Library]:
        pass

@runtime_checkable
class Writer_p(Protocol):
    """ Writers take a library, format it, and write it to a file.
    *typically* it formats as bibtex, but doesn't *have* to.
    (eg: RstWriter)
    """

    def write(self, library:Library, *, file:Maybe[pl.Path]=None, append:Maybe[list[Middleware]]=None) -> str:
        pass

##--| Middleware protocols

@runtime_checkable
class CustomWriteBlock_p(Protocol):
    """ Writers can be Visitors, in which case they call ths visit method on
    blocks

    """

    def visit(self, writer:Writer_p) -> list[str]:
        pass

@runtime_checkable
class ReadTime_p(Protocol):
    """ Protocol for signifying a middleware is for use on parsing bibtex to
    data
    """

    def on_read(self) -> Never:
        pass

@runtime_checkable
class WriteTime_p(Protocol):
    """ Protocol for signifying middleware is for use on writing data to bibtex

    """

    def on_write(self) -> Never:
        pass

@runtime_checkable
class EntrySkipper_p(Protocol):
    """ A whitelist based test for middlewares.
    Middleware's set their skiplist on init,
    and can call 'should_skip_entry' when transforming blocks

    eg: for only processing type='article' entries, not books
    """

    def set_entry_skiplist(self, whitelist:list[str]) -> None:
        pass

    def should_skip_entry(self, entry:Entry, library:Library) -> bool:
        pass

@runtime_checkable
class FieldMatcher_p(Protocol):
    """ The protocol util.FieldMatcher_m relies on

    A Middleware with the FieldMatcher_m mixin will call the implemented field_h
    on each field that matches in an entry.
    """

    def set_field_matchers(self, *, white:list[str], black:list[str]) -> Self:
        pass

    def match_on_fields(self, entry: Entry, library: Library) -> Result[Entry, Exception]:
        pass

    def field_h(self, field:Field, entry:Entry) -> Result[list[Field], Exception]:
        pass

@runtime_checkable
class StrTransformer_p(Protocol):
    """ Describes the StringTransform_m """

    def transform_strlike(self, slike:StrLike) -> Result[StrLike, Exception]:
        pass

    def _transform_raw_str(self, python_string:str) -> Result[str, Exception]:
        pass

@runtime_checkable
class DependentMiddleware_p(Protocol):
    """
    For middlewares that depend on a middleware to be able to work themselves.

    eg: metadata applicator requires path reader
    """

    def requires_in_same_stack(self) -> list[type]:
        """ The given types need to be in the same stack """
        pass

    def requires_in_parse_stack(self) -> list[type]:
        """ The given types need to be in the parse stack """
        pass



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)
from bibtexparser.middlewares.names import NameParts
from jgdv import Proto, Mixin
from jgdv.files.tags import SubstitutionFile

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from . import _interface as API_N
from bibble.util.mixins import ErrorRaiser_m, FieldMatcher_m
from bibble.fields.field_substitutor import FieldSubstitutor
from bibble.util.name_parts import NameParts_d

# ##-- end 1st party imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--|

@Proto(API.ReadTime_p)
class NameSubstitutor(FieldSubstitutor):
    """ replaces names in author and editor fields as necessary """

    def __init__(self, *, subs:SubstitutionFile, **kwargs):
        super().__init__(fields=[API_N.AUTHOR_K, API_N.EDITOR_K],
                         subs=subs,
                         **kwargs)

    def on_read(self):
        Never()

    def field_h(self, field, entry):
        match field.value:
            case str():
                return ValueError("Name parts should already be combined, but authors shouldn't be merged yet")
            case [*xs] if any(isinstance(x, NameParts|NameParts_d) for x in xs):
                return ValueError("Name parts should already be combined, but authors shouldn't be merged yet")
            case []:
                return [field]
            case [*xs]:
                clean_names = []
                for name in xs:
                    match self._subs.sub(name):
                        case None:
                            clean_names.append(name)
                        case set() as val:
                            head, *_ = val
                            clean_names.append(head)
                else:
                    return [model.Field(field.key, clean_names)]
            case value:
                return ValueError("Unsupported replacement field value type(%s): %s", entry.key, type(value))



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
import warnings

import pytest

from bibble.people.name_sub import NameSubstitutor
from bibtexparser import model
from jgdv.files.tags import SubstitutionFile

logging = logmod.root

class TestNameSubtitutor:

    @pytest.fixture(scope="function")
    def subber(self):
        """
        blah -> bloo
        von braun, Ernst -> von Braun, Ernst
        """
        subs = SubstitutionFile(norm_replace=" ", sep=" % ")
        subs.update("blah % 2 % bloo")
        subs.update("von braun, Ernst % 1 % von Braun, Ernst")
        return NameSubstitutor(subs=subs)

    @pytest.fixture(scope="function")
    def entry(self):
        return model.Entry("book", "test", [])

    ##--|
    def test_sanity(self, subber, entry):
        assert(True is True)
        assert("blah" in subber._subs)

    def test_no_value(self, subber, entry):
        entry.set_field(model.Field("author", []))
        match subber.transform_Entry(entry, None):
            case [x] if x is entry:
                assert(x.get("author").value == [])
            case x:
                 assert(False), x

    def test_no_sub(self, subber, entry):
        entry.set_field(model.Field("author", ["bob"]))
        match subber.transform_Entry(entry, None):
            case [x] if x is entry:
                assert(x.get("author").value == ["bob"])

    def test_simple_sub(self, subber, entry):
        entry.set_field(model.Field("author", ["blah"]))
        match subber.transform_Entry(entry, None):
            case [x] if x is entry:
                assert(x.get("author").value == ["bloo"])
            case x:
                 assert(False), x

    def test_selected_sub(self, subber, entry):
        entry.set_field(model.Field("author", ["jill", "blah", "bob"]))
        match subber.transform_Entry(entry, None):
            case [x] if x is entry:
                assert(x.get("author").value == ["jill", "bloo", "bob"])
            case x:
                assert(False), x

    def test_name_format_sub(self, subber, entry):
        entry.set_field(model.Field("author", ["von braun, Ernst"]))
        match subber.transform_Entry(entry, None):
            case [x] if x is entry:
                assert(x.get("author").value == ["von Braun, Ernst"])
            case x:
                 assert(False), x



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from jgdv import Proto, Mixin
import bibtexparser
import bibtexparser.model as model
import isbn_hyphenate
import pyisbn
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from . import _interface as MAPI
from bibble.util.middlecore import IdenBlockMiddleware
from bibble.util.mixins import ErrorRaiser_m
# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--|

@Proto(API.WriteTime_p)
@Mixin(ErrorRaiser_m)
class IsbnWriter(IdenBlockMiddleware):
    """
      format the isbn for writing
    """

    def on_write(self):
        Never()

    def transform_Entry(self, entry, library) -> list:
        f_dict = entry.fields_dict
        if MAPI.ISBN_K not in f_dict:
            return [entry]
        if MAPI.INVALID_ISBN_K in f_dict:
            return [entry]
        if not bool(f_dict[MAPI.ISBN_K].value):
            return [entry]

        try:
            isbn = isbn_hyphenate.hyphenate(f_dict[MAPI.ISBN_K].value)
            entry.set_field(model.Field(MAPI.ISBN_K, isbn))
            return [entry]
        except isbn_hyphenate.IsbnError as err:
            self._logger.warning("Writing ISBN failed: %s : %s", f_dict[MAPI.ISBN_K].value, err)
            return [entry,
                    self.make_error_block(entry, err)
                    ]



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from jgdv import Proto, Mixin
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.library import Library
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)
from jgdv.files.tags import SubstitutionFile

# ##-- end 3rd party imports

# ##-- 1st party imports
from bibble.model import MetaBlock
from bibble.util.mixins import ErrorRaiser_m, FieldMatcher_m
from bibble import _interface as API
from . import _interface as FieldsAPI
from bibble.util.middlecore import IdenBlockMiddleware

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Result
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Entry = model.Entry
    type Field = model.Field
    type Block = model.Block
    from bibtexparser.library import Library

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(API.FieldMatcher_p)
@Mixin(ErrorRaiser_m, FieldMatcher_m)
class FieldAccumulator(IdenBlockMiddleware):
    """ Create a set of all the values of a field, of all entries, in a library.

    'name' : the name of the accumulation block to store result in
    'fields' : the fields to accumulate values of.

    Fields can be individual values, or lists/sets of values

    """


    def __init__(self, *, name:str, fields:list[str], **kwargs):
        super().__init__(**kwargs)
        self._attr_target = name
        match fields:
            case list():
                self._target_fields = fields
            case x:
                raise TypeError(type(x))

        self.set_field_matchers(white=self._target_fields, black=[])
        self._collection = set()

    def transform(self, library:Library) -> Library:
        super().transform(library)
        library.add(FieldsAPI.AccumulationBlock(name=self._attr_target, data=self._collection, fields=self._target_fields))
        return library

    def transform_Entry(self, entry, library) -> list[Block]:
        match self.match_on_fields(entry, library):
            case model.Entry() as x:
                return [x]
            case Exception() as err:
                return [self.make_error_block(entry, err)]
            case x:
                raise TypeError(type(x))

    def field_h(self, field, entry) -> Result(list[Field], Exception):
        match field.value:
            case str() as value:
                self._collection.add(value)
            case list() | set() as value:
                self._collection.update(value)
            case x:
                raise TypeError(type(x))

        return []



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from random import choices
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)
from bibtexparser.middlewares.names import (NameParts,
                                            parse_single_name_into_parts)

# ##-- end 3rd party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable
    
    from bibtexparser.libary import Library

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class SelectN(LibraryMiddleware):
    """ Select N random entries """

    def __init__(self, *, count:int=1):
        super().__init__()
        self._count = count

    def transform(self, library):
        entries = library.entries
        chosen = choices(entries, k=self._count)
        return library

class SelectEntriesByType(LibraryMiddleware):
    """ Select entries of a particular type """
    default_targets = ("article",)

    def __init__(self, *, targets:list[str]):
        super().__init__()
        self._entry_target = target.lower()

    def transform(self, library):
        chosen = [x for x in library.entries if x.entry_type.lower() == self._entry_target]
        return Library(chosen)

class SelectTags(LibraryMiddleware):
    """ Select entries of with a particular tag """

    def __init__(self, *, tags:set[str]):
        super().__init__()
        self._targets = set(targets or [])

    def transform(self, library:Library) -> Library:
        chosen = [x for x in library.entries if bool(x.fields_dict['tags'].value & self._targets)]
        return Library(chosen)

class SelectAuthor(LibraryMiddleware):
    """ TODO select entries by a set of authors 
    
    should run name split on them
    """

    def __init__(self, *, authors:set[str]):
        super().__init__()
        self._targets = set(targets or [])

    def transform(self, library):
        raise NotImplementedError()



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from jgdv import Mixin, Proto
from jgdv.files.tags import TagFile

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from bibble.util.middlecore import IdenBlockMiddleware

# ##-- end 1st party imports

from . import _interface as MAPI

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from bibtexparser import Library

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(API.ReadTime_p)
class TagsReader(IdenBlockMiddleware):
    """
      Read Tag strings, split them into a set, and keep track of all mentioned tags
      By default the classvar _all_tags is cleared on init, pass clear=False to not

    tags are normalized through the TagFile
    """
    _all_tags : ClassVar[TagFile] = TagFile()

    _clear_on_transform : bool

    @staticmethod
    def tags_to_str():
        return str(TagsReader._all_tags)

    def __init__(self, *, clear:bool=True, **kwargs):
        super().__init__(**kwargs)
        self._clear_on_transform = clear

    def on_read(self):
        Never()

    def transform(self, library:Library) -> Library:
        if self._clear_on_transform:
            TagsReader._all_tags = TagFile()

        return super().transform(library)

    def transform_Entry(self, entry, library) -> list:
        tf = TagsReader._all_tags

        match entry.get(MAPI.TAGS_K):
            case None:
                self._logger.warning("Entry has no Tags on parse: %s", entry.key)
                entry.set_field(model.Field(MAPI.TAGS_K, set()))
            case model.Field(value=val) if not bool(val):
                self._logger.warning("Entry has no Tags on parse: %s", entry.key)
                entry.set_field(model.Field(MAPI.TAGS_K, set()))
            case model.Field(value=str() as val):
                as_set = set(tf.norm_tag(x) for x in val.split(","))
                entry.set_field(model.Field(MAPI.TAGS_K, as_set))
                tf.update(as_set)
            case model.Field(value=set()):
                pass

        return [entry]



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

import bibble._interface as API

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Fifo, Lifo
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type UniMiddleware   = API.UniMiddleware
    type BidiMiddleware  = API.BidiMiddleware
    type Middleware      = API.Middleware

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class PairStack:
    """ A pair of middleware stacks,
    allowing reader/writer pairs to be added at the same time

    The read stack is Fifo, the write stack is Lifo.
    So any transforms are undone in the correct order before writing.
    Parse[m1, m2, m3] -> Write[m3, m2, m1]

    eg: Parse[LatexToUnicode, SplitName] -> Write[MergeNames, UnicodeToLatex]

    """
    _read_time  : Fifo[Middleware]
    _write_time : Lifo[Middleware]

    def __init__(self):
        self._read_time  = []
        self._write_time = []

    def add(self, *bidis:BidiMiddleware, read:Maybe[list[Middleware]]=None, write:Maybe[list[Middleware]]=None) -> Self:
        """
        Add middlewares to the read/write stacks.
        Args *must* be bidirection, and will be added to both.
        kwargs 'read' and 'write' just add to specific stacks
        """
        read  = read or []
        write = write or []

        for mid in bidis:
            match mid:
                case None:
                     pass
                case API.BidirectionalMiddleware_p():
                    self._read_time.append(mid)
                    self._write_time.append(mid)
                case x:
                    raise TypeError(type(x))

        for x in read:
            match x:
                case None:
                    pass
                case API.Middleware_p():
                    self._read_time.append(x)
                case API.BidirectionalMiddleware_p():
                    self._read_time.append(x)
                case _:
                    raise TypeError(type(x))

        for x in write:
            match x:
                case None:
                    pass
                case API.Middleware_p():
                    self._write_time.append(x)
                case API.BidirectionalMiddleware_p():
                    self._write_time.append(x)
                case _:
                    raise TypeError(type(x))

        return self

    def read_stack(self) -> list[Middleware]:
        """ Return the read stack """
        return self._read_time[:]

    def write_stack(self) -> list[Middleware]:
        """ Return the write stack """
        return self._write_time[::-1]

    def has_read_transform(self, mid:type[Middleware]|Middleware) -> bool:
        match mid:
            case type():
                return any(isinstance(x, mid) for x in self._read_time)
            case API.Middleware_p():
                return mid in self._read_time
            case _:
                return False

    def has_write_transform(self, mid:type[Middleware]|Middleware) -> bool:
        match mid:
            case type():
                return any(isinstance(x, mid) for x in self._write_time)
            case API.Middleware_p():
                return mid in self._write_time
            case _:
                return False

    def __contains__(self, other:type[Middleware]|Middleware) -> bool:
        return self.has_read_transform(other) or self.has_write_transform(other)



#!/usr/bin/env python3
"""

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import base64
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

from selenium.webdriver import Firefox, FirefoxOptions, FirefoxService
from selenium.webdriver.common.print_page_options import PrintOptions
import bibble._interface as API
from . import _interface as FAPI
# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class FirefoxController:
    """ A Static controller for starting and closing firefox via selenium """

    @staticmethod
    def setup(*, opts:list, kwargs:dict) -> None:
        """ Setups a selenium driven, headless firefox to print to pdf

        """
        if hasattr(FirefoxController, FAPI.FF_DRIVER):
            logging.info("Skipping Firefox Setup")
            return getattr(FirefoxController, FAPI.FF_DRIVER)

        logging.info("Setting up headless Firefox")
        options = FirefoxOptions()
        for x in opts:
            options.add_argument(x)

        for x,y in kwargs.items():
            options.set_preference(x, y)

        # options.binary_location = "/usr/bin/firefox"
        # options.binary_location = "/snap/bin/geckodriver"
        service                  = FirefoxService(executable_path=FAPI.GECKO_DRIVER)
        driver                   = Firefox(options=options, service=service)
        driver.set_page_load_timeout(FAPI.LOAD_TIMEOUT)
        setattr(FirefoxController, FAPI.FF_DRIVER, driver)
        return driver

    @staticmethod
    def close() -> None:
        if not hasattr(FirefoxController, FAPI.FF_DRIVER):
            return

        logging.info("Closing Firefox")
        getattr(FirefoxController, FAPI.FF_DRIVER).quit()

    @staticmethod
    def save_pdf(url, dest) -> None:
        """ prints a url to a pdf file using selenium """
        if not isinstance(dest, pl.Path):
            raise FileNotFoundError("Destination to save pdf to is not a path", dest)

        if dest.suffix != ".pdf":
            raise FileNotFoundError("Destination isn't a pdf", dest)

        if dest.exists():
            logging.info("Destination already exists: %s", dest)
            return

        driver = FirefoxController.setup(opts=FAPI.SELENIUM_OPTS, kwargs=FAPI.SELENIUM_PREFS)
        logging.info("Saving: %s", url)
        print_ops = PrintOptions()
        print_ops.page_range = "all"

        driver.get(FAPI.READER_PREFIX + url)
        time.sleep(FAPI.LOAD_TIMEOUT)
        pdf       = driver.print_page(print_options=print_ops)
        pdf_bytes = base64.b64decode(pdf)

        if not bool(pdf_bytes):
            logging.warning("No Bytes were downloaded")
            return

        logging.info("Saving to: %s", dest)
        with dest.open("wb") as f:
            f.write(pdf_bytes)



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from jgdv import Proto, Mixin
import bibtexparser
import bibtexparser.model as model
from bibtexparser.library import Library
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware, LibraryMiddleware)

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from bibble.util.mixins import ErrorRaiser_m, FieldMatcher_m
from bibble.util.middlecore import IdenBlockMiddleware
from bibble.model import MetaBlock

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Entry = model.Entry
    type Field = model.Field
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@Proto(API.WriteTime_p)
@Mixin(ErrorRaiser_m, FieldMatcher_m)
class PathWriter(IdenBlockMiddleware):
    """
      Relativize library paths back to strings

    Can suppress errors from certain path roots on relativize,
    using MetaBlock data:
    MetaBlock(PathWriter.SuppressKey=[pl.Path()...])

    """

    _whitelist  = ("file",)
    SuppressKey = "PathWriter.suppress"
    _suppress_in : list[pl.Path]

    def __init__(self, *, lib_root:Maybe[pl.Path]=None, **kwargs):
        super().__init__(**kwargs)
        self.set_field_matchers(white=self._whitelist, black=[])
        self._lib_root    = lib_root
        self._suppress_in = []

    def handle_meta_entry(self, library:Library):
        match MetaBlock.find_in(library):
            case None:
                return
            case block if PathWriter.SuppressKey not in block.data:
                return
            case block:
                pass

        match block.data[PathWriter.SuppressKey]:
            case list() as xs:
                self._suppress_in += xs
                return
            case x:
                raise TypeError(f"{PathWriter.SuppressKey} is not a list, but a {type(x)}")

    def on_write(self):
        Never()

    def transform_Entry(self, entry:Entry, library:Library):
        match self.match_on_fields(entry, library):
            case model.Entry() as x:
                return [x]
            case Exception() as err:
                return [entry, self.make_error_block(entry, err)]
            case x:
                raise TypeError(type(x), x)

    def field_h(self, field:Field, entry:Entry):
        match field.value:
            case str():
                pass
            case pl.Path() as val if not val.exists():
                return ValueError(f"On Export file does not exist: {entry.key} : {val}")
            case pl.Path() as val:
                try:
                    as_str = val.relative_to(self._lib_root)
                    field.value = as_str
                except ValueError:
                    if self._suppress_relative_fail(val):
                        pass
                    else:
                        field.value = str(val)
                        return ValueError(f"Failed to Relativize path {entry.key}: {val}")

        return [field]

    def _suppress_relative_fail(self, val:pl.Path) -> bool:
        for x in self._suppress_in:
            try:
                val.relative_to(x)
                return True
            except ValueError:
                pass
        else:
            return False



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import json
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser as BTP
import jsonlines
import sh
from bibtexparser import model
from bibtexparser.middlewares import NameParts

from jgdv import Proto, Mixin

# ##-- end 3rd party imports

import bibble._interface as API
from . import _interface as MAPI
from bibble.util.middlecore import IdenBlockMiddleware
from bibble.util.mixins import ErrorRaiser_m
from bibble.util.name_parts import NameParts_d

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Field  = model.Field
    type Entry  = model.Entry
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

exiftool = sh.exiftool
calibre  = sh.ebook_meta
qpdf     = sh.qpdf
pdfinfo  = sh.pdfinfo
##--|

class _EntryFileGetter_m:
    """ Mixin for getting the paths of files in the entry
    can return a list of all file_{N} field values
    """

    def _get_files(self, entry) -> list[pl.Path]:
        """ gets all paths of fields with 'file' in the field name """
        paths = []
        for field in entry.fields:
            match field.value:
                case _ if MAPI.FILE_K not in field.key:
                    continue
                case None:
                    continue
                case pl.Path() as path:
                    paths.append(paths)
                case str() as pathstr:
                    paths.append(pl.Path(pathstr))
                case x:
                    raise TypeError(type(x))
        else:
            return paths

    def _get_file(self, entry) -> Maybe[pl.Path]:
        """ Gets the main file fields path of an entry """
        match entry.fields_dict.get(MAPI.FILE_K, None):
            case BTP.model.Field(value=pl.Path() as path):
                return path
            case BTP.model.Field(value=str() as pathstr):
                return pl.Path(pathstr)
            case _:
                return None

class _Metadata_Check_m:
    """A mixin for checking the metadata fof files"""

    def backup_original_metadata(self, path:pl.Path) -> None:
        """
        If self._backup is set, backup the files metadata as jsonlines there
        Uses exiftool to export the metadata as json,
        which is then appended into the backup as a jsonlines file.
        """
        match self._backup:
            case pl.Path():
                pass
            case _:
                return

        try:
            result = json.loads(exiftool("-J", str(path)))[0]
        except sh.ErrorReturnCode:
            raise ValueError("Couldn't retrieve metadata as json", path) from None

        with jsonlines.open(self._backup, mode='a') as f:
            f.write(result)

    def metadata_matches_entry(self, path:pl.Path, entry:Entry) -> bool:
        """ Test the given path to see if the metadata matches.
        This is quite naive.
        From exiftool, it looks for either:
        - a 'Bibtex' field,
        - a 'Description' field,

        and compares that to the raw entry's text.

        TODO switch to use a stored hash instead
        """
        try:
            result = json.loads(exiftool("-J", str(path)))[0]
        except sh.ErrorReturnCode:
            self._logger.warning("Couldn't match metadata", path)
            return False

        bib_field_matches  = result.get(MAPI.BIBTEX_EXIF, None) == entry.raw
        desc_field_matches = result.get(MAPI.DESC_EXIF, None) == entry.raw

        return bib_field_matches or desc_field_matches

class _Pdf_Update_m:
    """A Mixin for pdf specific metadata manipulation"""

    def update_pdf_by_exiftool(self, path:pl.Path, entry:Entry) -> None:
        """
        exiftool -{tag}="{content}" {file}
        https://exiftool.org/
        """
        # Build args:
        args = self._entry_to_exiftool_args(entry)

        self._logger.debug("Pdf update args: %s : %s", path, args)
        # Call
        try:
            exiftool(*args, str(path))
        except sh.ErrorReturnCode as err:
            raise ChildProcessError("Exiftool update failed", err) from None

    def pdf_is_modifiable(self, path:pl.Path) -> bool:
        """ Use qpdf to test the pdf for encryption or password locking,
        """
        try:
            cmd1 = qpdf(MAPI.QPDF_IS_ENCRPT, str(path), _ok_code=MAPI.QPDF_OK_CODES)
            cmd2 = qpdf(MAPI.QPDF_REQ_PASS, str(path), _ok_code=MAPI.QPDF_OK_CODES)
        except sh.ErrorReturnCode as err:
            return False

        return True

    def pdf_validate(self, path:pl.Path) -> None:
        """
        Validates a pdf using qpdf
        https://qpdf.readthedocs.io/en/stable/

        code 0 for fine,
        code 2 for errors
        code 3 for warnings
        writes to stderr for issues
        """
        try:
            qpdf(MAPI.QPDF_CHECK, str(path))
        except sh.ErrorReturnCode:
            raise ChildProcessError("PDF Failed Validation") from None

    def pdf_finalize(self, path:pl.Path) -> None:
        """ run qpdf --linearize,

        on success, delete the original if it exists
        """
        assert(path.suffix == MAPI.PDF_SUFF)
        self._logger.debug("Finalizing Pdf: %s", path)
        original = str(path)
        copied   = path.with_stem(path.stem + MAPI.PDF_COPY_SUFF)
        backup   = path.with_suffix(MAPI.PDF_ORIG_SUFF)
        if copied.exists():
            raise FileExistsError("The temp copy for linearization shouldn't already exist", original)

        path.rename(copied)
        try:
            qpdf(str(copied), MAPI.QPDF_LINEAR, original)
        except sh.ErrorReturnCode:
            copied.rename(original)
            raise ChildProcessError("Linearization Failed") from None
        else:
            if backup.exists():
                backup.unlink()
            if original.exists() and copied.exists() and original != copied:
                copied.unlink()

    def _entry_to_exiftool_args(self, entry) -> list[str]:
        """
        Extract and format the entry as exiftool args
        Uses XMP for the metadata under a custom XMP-bib namespace
        # TODO add exiftool config to bibble
        """
        fields = entry.fields_dict
        args   = []

        # XMP-bib:
        # The custom full bibtex entry
        args += [f'-bibtex={entry.raw}']

        # General
        match fields:
            case {"title": t, "subtitle": st}:
                args += ['-title={}: {}'.format(t.value, st.value)]
            case {"title": t}:
                args += ['-title={}'.format(t.value)]
            case {"short_parties": t}:
                args += ['-title={}'.format(t.value)]

        match fields:
            case {"author": a}:
                args += ['-author={}'.format(a.value)]
            case {"editor": a}:
                args += ['-author={}'.format(a.value)]

        args += ['-Year={}'.format(fields['year'].value)]

        if 'tags' in fields:
            args += ['-Keywords={}'.format(",".join(fields['tags'].value))]
        if 'isbn' in fields:
            args += ['-ISBN={}'.format(fields['isbn'].value)]
        if 'edition' in fields:
            args += ['-xmp-prism:edition={}'.format(fields['edition'].value)]
        if 'publisher' in fields:
            args += ["-publisher={}".format(fields['publisher'].value)]
        if 'url' in fields:
            args += ["-xmp-prism:link={}".format(fields['url'].value)]
        if 'doi' in fields:
            args += ['-xmp-prism:DOI={}'.format(fields['doi'].value)]
        if 'institution' in fields:
            args += ['-xmp-prism:organization={}'.format(fields['institution'].value)]
        if 'issn' in fields:
            args += ['-xmp-prism:issn={}'.format(fields['issn'].value)]

        return args

class _Epub_Update_m:
    """A Mixin for epub-specific metadata manipulation"""

    def update_epub_by_calibre(self, path, entry) -> None:
        """ Uses calibre to modify epub metadata
        https://manual.calibre-ebook.com/generated/en/cli-index.html
        """
        args = self.entry_to_calibre_args(entry)

        self._logger.debug("Ebook update args: %s : %s", path, args)
        try:
            calibre(str(path), *args)
        except sh.ErrorReturnCode:
            raise ChildProcessError("Calibre Update Failed") from None

    def entry_to_calibre_args(self, entry) -> list[str]:
        fields = entry.fields_dict
        args   = []
        title  = None

        match fields:
            case {"title":t, "subtitle":st}:
                args.append(f"--title={t.value}: {st.value}")
            case {"title":t}:
                args.append(f"--title={t.value}")
            case {"short_parties":t}:
                args.append(f"--title={t.value}")

        # authors should be joined into one string already
        match fields:
            case {"author":str() as auth}:
                args += ['--authors={}'.format(auth.value)]
            case {"editor":str() as ed}:
                args += ['--authors={}'.format(ed.value)]
            case {"author":list()|NameParts()|NameParts_d() as bad}:
                raise TypeError("Author Should have been reduced already", entry.key, bad)
            case {"editor":list()|NameParts()|NameParts_d() as bad}:
                raise TypeError("Editor Should have been reduced already", entry.key, bad)

        if 'publisher' in fields:
            args += ["--publisher={}".format(fields['publisher'].value)]
        if 'series' in fields:
            args += ["--series={}".format(fields['series'].value)]
        if 'number' in fields:
            args += ['--index={}'.format(fields['number'].value)]
        if 'volume' in fields:
            args += ['--index={}'.format(fields['volume'].value)]
        if 'isbn' in fields:
            args += ['--isbn={}'.format(fields['isbn'].value)]
        if 'doi' in fields:
            args += ['--identifier=doi:{}'.format(fields['doi'].value)]
        if 'tags' in fields:
            args += ['--tags={}'.format(",".join(fields['tags'].value))]
        if 'year' in fields:
            args += ['--date={}'.format(fields['year'].value)]

        args += ['--comments={}'.format(entry.raw)]

        return args

##--|

@Proto(API.WriteTime_p)
@Mixin(_Pdf_Update_m, _Epub_Update_m, _EntryFileGetter_m, _Metadata_Check_m, ErrorRaiser_m)
class ApplyMetadata(IdenBlockMiddleware):
    """ Apply metadata to files mentioned in bibtex entries
      uses xmp-prism tags and some custom ones for pdfs,
      and epub standard.

    TODO add a 'meta_update' status field to the entry for [locked,failed]
      """
    _backup   : Maybe[pl.path]
    _failures : list[Exception]

    def __init__(self, *, backup:Maybe[pl.Path]=None, **kwargs):
        super().__init__(**kwargs)
        self._extra.setdefault("tqdm", True)
        self._backup   = backup
        self._failures = []

    def on_write(self):
        Never()

    def transform_Entry(self, entry,  library) -> Any:
        result = []
        match self._get_file(entry):
            case None:
                pass
            case pl.Path() as x if not x.exists():
                update = BTP.model.Field(MAPI.ORPHANED_K, True)
                entry.set_field(update)
                result.append(entry)
            case pl.Path() as x if self.metadata_matches_entry(x, entry):
                self._logger.info("No Metadata Update Necessary: %s", x)
            case pl.Path() as x if x.suffix == MAPI.PDF_SUFF:
                for field in self.process_pdf(x, entry):
                    entry.set_field(field)
                else:
                    result.append(entry)
            case pl.Path() as x if x.suffix == MAPI.EPUB_SUFF:
                for field in self.process_epub(x, entry):
                    entry.set_field(field)
                else:
                    result.append(entry)
            case x:
                self._failures.append(TypeError("Unknown File Type", entry.key, x))
                self._logger.warning("Found a file that wasn't an epub or pdf: %s", x)

        for x in self._failures:
            result.append(self.make_error_block(entry, x))
        else:
            self._failures = []
            return result

    def process_epub(self, epub:pl.Path(), entry) -> list[Field]:
        try:
            self.backup_original_metadata(epub)
            self.update_epub_by_calibre(epub, entry)
        except (ValueError, ChildProcessError) as err:
            self._failures.append(ValueError("Epub meta update failed", epub, *err.args))
            self._logger.warning("Epub Update failed: %s : %s", epub, err)
        else:
            if not epub.exists():
                raise FileNotFoundError("File has gone missing", epub)

    def process_pdf(self, pdf:pl.Path(), entry) -> list[Field]:
        if not self.pdf_is_modifiable(pdf):
            locked_field = BTP.model.Field(MAPI.PDF_LOCKED_K , True)
            self._failures.append(ValueError("Pdf is locked", pdf))
            return [locked_field]

        try:
            self.backup_original_metadata(pdf)
            self.update_pdf_by_exiftool(pdf, entry)
            self.pdf_validate(pdf)
            self.pdf_finalize(pdf)
        except (ValueError, ChildProcessError, FileExistsError) as err:
            self._failures.append(ValueError("Pdf Meta update failed", pdf, *err.args))
            self._logger.warning("Pdf Update Failed: %s : %s", pdf, err)
            return []
        else:
            if not pdf.exists():
                raise FileNotFoundError("File has gone missing", pdf)

##--|

@Mixin(_Pdf_Update_m, _EntryFileGetter_m)
class FileCheck(IdenBlockMiddleware):
    """ Like ApplyMetadata, but just checks for files that can't be modified or are missing,
    so they can be fixed.
    ie: its faster

      Annotate entries with 'pdf_locked' if the pdf can't be modified,
      "orphan_file" if the pdf or epub does not exist
    """

    def transform_Entry(self, entry, library) -> list:
        """
        TODO remove orphan/lock field if its no longer the case
        """
        match self._get_file(entry):
            case None:
                return []
            case pl.Path() as x if not x.exists():
                update = BTP.model.Field(MAPI.ORPHANED_K, True)
                entry.set_field(update)
            case pl.Path() as x if x.suffix == MAPI.PDF_SUFF and MAPI.PDF_LOCKED_K in entry.fields_dict:
                pass
            case pl.Path() as x if x.suffix == MAPI.PDF_SUFF and not self.pdf_is_modifiable(x):
                update = BTP.model.Field(MAPI.PDF_LOCKED_K, True)
                entry.set_field(update)
            case pl.Path() as x if x.suffix == MAPI.PDF_SUFF:
                update = BTP.model.Field(MAPI.PDF_LOCKED_K,  False)
                entry.set_field(update)
            case _:
                return []

        return [entry]



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

import bibble._interface as API
from bibtexparser import Library
from .. import Reader
from bibble.bidi import BraceWrapper

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
EXAMPLE_BIB : Final[str] = """
@article{test_art,
  title  = {Blah},
  year   = {1992},
  author = {Bob},
}

"""
# Body:

class TestBibbleReader:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match Reader([]):
            case API.Reader_p():
                assert(True)
            case x:
                 assert(False), x

    def test_read(self):
        reader = Reader([])
        match reader.read(EXAMPLE_BIB):
            case Library() as lib:
                assert(len(lib.entries) == 1)
                entry                                    = lib.entries[0]
                assert(entry.key                         == "test_art")
                assert(entry.entry_type                  == "article")
                assert(entry.fields_dict['title'].value  == "{Blah}")
                assert(entry.fields_dict['author'].value == "{Bob}")
                assert(entry.fields_dict['year'].value   == "{1992}")
            case x:
                 assert(False), x


    def test_read_with_middleware(self):
        reader = Reader([BraceWrapper()])
        match reader.read(EXAMPLE_BIB):
            case Library() as lib:
                assert(len(lib.entries) == 1)
                entry                                    = lib.entries[0]
                assert(entry.key                         == "test_art")
                assert(entry.entry_type                  == "article")
                assert(entry.fields_dict['title'].value  == "Blah")
                assert(entry.fields_dict['author'].value == "Bob")
                assert(entry.fields_dict['year'].value   == "1992")
            case x:
                 assert(False), x

    @pytest.mark.skip
    def test_todo(self):
        pass



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

from bibtexparser import Library, model
import bibble._interface as API
from .. import NameReader
from ..name_reader import _SplitAuthors_m, _NameToParts_m, NameParts_d

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class TestSplitAuthors:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic_split(self):
        obj = _SplitAuthors_m()
        match obj._split_authors("bob and jim"):
            case ["bob", "jim"]:
                assert(True)
            case x:
                 assert(False), x

    def test_no_split(self):
        obj = _SplitAuthors_m()
        match obj._split_authors("{bob and jim}"):
            case ["{bob and jim}"]:
                assert(True)
            case x:
                 assert(False), x

    def test_split_some(self):
        obj = _SplitAuthors_m()
        match obj._split_authors("andy and {bob and jim} and jill"):
            case ["andy", "{bob and jim}", "jill"]:
                assert(True)
            case x:
                 assert(False), x

    def test_split_nothing(self):
        obj = _SplitAuthors_m()
        match obj._split_authors(""):
            case []:
                assert(True)
            case x:
                assert(False), x

class TestNameToParts:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        obj = _NameToParts_m()
        match obj._name_to_parts("Bob"):
            case NameParts_d() as np:
                assert(np.last == ["Bob"])
            case x:
                 assert(False), x

    def test_first_last(self):
        obj = _NameToParts_m()
        match obj._name_to_parts("Bob Builder"):
            case NameParts_d() as np:
                assert(np.first == ["Bob"])
                assert(np.last == ["Builder"])
            case x:
                 assert(False), x

    def test_first_initials_last(self):
        obj = _NameToParts_m()
        match obj._name_to_parts("Bob A. B. C. Builder"):
            case NameParts_d() as np:
                assert(np.first == ["Bob", "A.", "B.", "C."])
                assert(np.last == ["Builder"])
            case x:
                 assert(False), x

    def test_first_von_last(self):
        obj = _NameToParts_m()
        match obj._name_to_parts("Bob von Builder"):
            case NameParts_d() as np:
                assert(np.first == ["Bob"])
                assert(np.last == ["Builder"])
                assert(np.von == ["von"])
            case x:
                 assert(False), x

    def test_first_de_la_last(self):
        obj = _NameToParts_m()
        match obj._name_to_parts("Bob de la Builder"):
            case NameParts_d() as np:
                assert(np.first == ["Bob"])
                assert(np.last == ["Builder"])
                assert(np.von == ["de", "la"])
            case x:
                 assert(False), x

    def test_last_first(self):
        obj = _NameToParts_m()
        match obj._name_to_parts("Builder, Bob"):
            case NameParts_d() as np:
                assert(np.first == ["Bob"])
                assert(np.last == ["Builder"])
            case x:
                 assert(False), x

    def test_von_last_first(self):
        obj = _NameToParts_m()
        match obj._name_to_parts("von Builder, Bob"):
            case NameParts_d() as np:
                assert(np.first == ["Bob"])
                assert(np.last == ["Builder"])
                assert(np.von == ["von"])
            case x:
                 assert(False), x

    def test_von_last_jr_first(self):
        obj = _NameToParts_m()
        match obj._name_to_parts("von Builder, jr, Bob"):
            case NameParts_d() as np:
                assert(np.first == ["Bob"])
                assert(np.last == ["Builder"])
                assert(np.von == ["von"])
                assert(np.jr == ["jr"])
            case x:
                 assert(False), x

    def test_last_first_initials(self):
        obj = _NameToParts_m()
        match obj._name_to_parts("Builder, Bob A. B. C."):
            case NameParts_d() as np:
                assert(np.first == ["Bob", "A.", "B.", "C."])
                assert(np.last == ["Builder"])
            case x:
                 assert(False), x

    def test_wrapped(self):
        obj = _NameToParts_m()
        match obj._name_to_parts("{The Order}"):
            case NameParts_d() as np:
                assert(np.last == ["{The Order}"])
            case x:
                 assert(False), x

class TestNameReader:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match NameReader():
            case API.Middleware_p():
                assert(True)
            case x:
                 assert(False), x

    def test_split_names(self):
        authors = model.Field("author", "Bill and Bob")
        entry   = model.Entry("test", "test:blah", [authors])
        lib     = Library([entry])
        mid     = NameReader(authors=True, parts=False)
        match mid.transform(lib):
            case Library() as l2:
                assert(l2 is lib)
                assert(not bool(lib.failed_blocks))
                assert(l2.entries[0].fields[0].value == ["Bill", "Bob"])
            case x:
                 assert(False), x

    def test_name_parts(self):
        authors = model.Field("author", "Bill and Bob")
        entry   = model.Entry("test", "test:blah", [authors])
        lib     = Library([entry])
        mid     = NameReader(authors=True, parts=True)
        match mid.transform(lib):
            case Library() as l2:
                assert(l2 is lib)
                assert(not bool(lib.failed_blocks))
                assert(len(l2.entries[0].fields[0].value) == 2)
            case x:
                 assert(False), x

        match l2.entries[0].fields[0].value:
            case [NameParts_d() as a, NameParts_d() as b]:
                assert(a.last == ["Bill"])
                assert(b.last == ["Bob"])
            case x:
                 assert(False), x

    @pytest.mark.skip
    def test_todo(self):
        pass



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

from importlib.resources import files
import bibble as BM
import bibtexparser as BTP
from bibtexparser import Library
from bibtexparser import middlewares as ms
from bibble.util import PairStack

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##-- data
data_path : pl.Path = files(__package__) / "_data"
##-- end data

# Vars:
sort_firsts = ["title", "author", "editor", "year", "tags", "booktitle", "journal", "volume", "number", "edition", "edition_year", "publisher"]
sort_lasts  = ["isbn", "doi", "url", "file", "crossref"]
sub_fields  = ["publisher", "journal", "series", "institution"]
meta_keys   = {"all", "apply", "latex", "rst", "subs", "check", "fsort", "esort", "count",}
# Body:

def build_new_stack(_libroot:pl.Path, *, _namesubs:Maybe=None, _tagsubs:Maybe=None, _othersubs:Maybe=None, **kwargs) -> PairStack:
    """ Build a new PairStack of middlewares, with optional and required elements
    Because of how pairstack works, to see the parse stack, read from top to bottom.
    To see the write transforms, read from bottom to top.
    """
    _meta = set(x for x,y in kwargs.items() if y is True)
    if bool((extra:=_meta - meta_keys)):
        msg = "Unrecognised meta keys provided"
        raise ValueError(msg, extra)

    ALL   = "all" in _meta
    APPLY = ALL or "apply" in _meta
    LATEX = ALL or "latex" in _meta
    FSORT = ALL or "fsort" in _meta
    ESORT = ALL or "esort" in _meta
    COUNT = ALL or "count" in _meta
    SUBS  = ALL or "subs" in _meta
    CHECK = ALL or "check" in _meta

    stack = PairStack()
    # Very first/last middlewares:
    #
    stack.add(read=[BM.failure.DuplicateKeyHandler(),
                    ],
              write=[
                  BM.failure.FailureHandler(),
                  BM.metadata.ApplyMetadata() if APPLY else None,
              ])
    # Add bidirectional transforms
    stack.add(BM.bidi.BraceWrapper(),
              BM.bidi.BidiLatex() if LATEX else None,
              BM.bidi.BidiPaths(lib_root=_libroot),
              BM.bidi.BidiNames(parts=True, authors=True),
              BM.bidi.BidiIsbn(),
              BM.bidi.BidiTags(),
              None,
              read=[
                  BM.metadata.KeyLocker(),
                  BM.fields.TitleSplitter()
              ],
              write=[
                  BM.fields.FieldSorter(first=sort_firsts, last=sort_lasts) if FSORT else None,
                  BM.metadata.EntrySorter() if ESORT else None,
              ])

    if COUNT:
        # Accumulate various fields
        stack.add(write=[
            BM.fields.FieldAccumulator(name="all-tags",     fields=["tags"]),
            BM.fields.FieldAccumulator(name="all-pubs",     fields=["publisher"]),
            BM.fields.FieldAccumulator(name="all-series",   fields=["series"]),
            BM.fields.FieldAccumulator(name="all-journals", fields=["journal"]),
            BM.fields.FieldAccumulator(name="all-people",   fields=["author", "editor"]),
        ])

    if SUBS:
        stack.add(write=[
            # NameSubs need to merge with BidiNames
            # BM.people.NameSubstitutor(_namesubs) if _namesubs is not None else None,
            BM.fields.FieldSubstitutor(fields=["tags"], subs=_tagsubs) if _tagsubs is not None else None,
            BM.fields.FieldSubstitutor(fields=sub_fields, subs=_othersubs, force_single_value=True) if _othersubs is not None else None,
        ])

    if CHECK:
        stack.add(write=[BM.metadata.FileCheck()])

    stack.add(read=[BM.failure.FailureHandler()])
    return stack

##--|

class TestFullStack:

    def test_sanity(self):
        assert(True is not False)

    def test_simple_stack(self):
        match build_new_stack(pl.Path.cwd()):
            case PairStack():
                assert(True)
            case x:
                 assert(False), x

    def test_bad_meta_val(self):
        with pytest.raises(ValueError):
            build_new_stack(pl.Path.cwd(), unknown=True)

    def test_apply_kw(self):
        without_stack = build_new_stack(pl.Path.cwd(), apply=False)
        with_stack    = build_new_stack(pl.Path.cwd(), apply=True)

        assert(BM.metadata.ApplyMetadata not in without_stack)
        assert(BM.metadata.ApplyMetadata in with_stack)


    def test_full_reader(self):
        stack = build_new_stack(pl.Path.cwd(), all=True)
        reader = BM.io.Reader(stack)
        assert(True)

    def test_full_writer(self):
        stack = build_new_stack(pl.Path.cwd(), all=True)
        reader = BM.io.Reader(stack)
        assert(True)

    def test_read(self):
        stack = build_new_stack(pl.Path.cwd())
        reader = BM.io.Reader(stack)
        match reader.read(data_path / "1320_orig.bib"):
            case Library() as lib:
                assert(bool(lib.entries))
                assert(not bool(lib.failed_blocks))
                assert(len(lib.entries) == 3)
                authors = {repr(x.fields_dict['author'].value) for x in lib.entries}
                assert(len(authors) == 1)
            case x:
                assert(False), x


    def test_write(self):
        stack       = build_new_stack(pl.Path.cwd())
        reader      = BM.io.Reader(stack)
        lib         = reader.read(data_path / "1320_orig.bib")
        cleaned     = (data_path / "1320_clean.bib").read_text().strip()
        writer      = BM.io.Writer(stack)
        match writer.write(lib):
            case str() as x:
                assert(bool(x))
                assert(x.strip() == cleaned)
            case x:
                assert(False), x


    def test_read_merge(self):
        stack        = build_new_stack(pl.Path.cwd())
        reader       = BM.io.Reader(stack)
        lib          = reader.read(data_path / "1320_orig.bib")
        starting_len = len(lib.entries)
        match reader.read(data_path / "1868.bib", into=lib):
            case Library() as lib2:
                assert(starting_len < len(lib2.entries))
                assert(lib2 is lib)
            case x:
                 assert(False), x


    def test_read_failure(self, caplog):
        """ reads a file with two entries with the same key """
        stack        = build_new_stack(pl.Path.cwd())
        reader       = BM.io.Reader(stack)
        match reader.read(data_path / "1320_duplicates.bib"):
            case Library():
                assert("Handling 1 failed blocks" in caplog.text)
                assert("Adjusted 1 duplicate keys" in caplog.text)
                assert(True)
            case x:
                assert(False), x



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser.middlewares.middleware import (BlockMiddleware, LibraryMiddleware)
from jgdv import Proto, Mixin
from jgdv.files.bookmarks import BookmarkCollection
from jgdv.files.tags import TagFile
from waybackpy import WaybackMachineSaveAPI

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from . import  _interface as FAPI
from ._firefox import FirefoxController
from bibble.util.mixins import FieldMatcher_m, EntrySkipper_m
from bibble.util.middlecore import IdenBlockMiddleware

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from bibtexparser.library import Library

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--|

@Mixin(EntrySkipper_m)
class OnlineDownloader(IdenBlockMiddleware):
    """
      if the entry is 'online', and it doesn't have a file associated with it,
      download it as a pdf and add it to the entry
    """
    _whitelist = ("online", "blog")
    _target_dir         : pl.Path

    def __init__(self, *, target:pl.Path, **kwargs):
        super().__init__(**kwargs)
        self._extra.setdefault("tqdm", True)
        self.set_entry_skiplists(white=self._whitelist, black=[])
        self._target_dir         = target

    def transform(self, library:Library) -> Library:
        try:
            result = super().transform(library)
            return result
        finally:
            FirefoxController.close()

    def transform_Entry(self, entry, library):
        if self.should_skip_entry(entry, library):
            return [entry]

        match entry.get("url"), entry.get("file"):
            case _, pl.Path()|str():
                self._logger.info("Entry %s : Already has file", entry.key)
                return entry
            case None, _:
                self._logger.warning("Entry %s : no url found", entry.key)
                return entry
            case model.Field(value=url), None:
                safe_key = entry.key.replace(":","_")
                dest     = (self._target_dir / safe_key).with_suffix(".pdf")
                FirefoxController.save_pdf(url, dest)
                # add it to the entry
                entry.set_field(model.Field("file", value=dest))

        return [entry]



#!/usr/bin/env python3
"""
Refactored from bibtexparser.middlewares.names

"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe, M_
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type MbList = Maybe[list[str]]
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
JOIN_STR : Final[str] = ", "
# Body:

def escape_last_slash(string: str) -> str:
    """Escape the last slash in a string if it is not escaped."""
    # Find the number of trailing slashes
    stripped = string.rstrip("\\")
    num_slashes = len(string) - len(stripped)
    if num_slashes % 2 == 0:
        # Even number: everything is escaped
        return string
    else:
        # Odd number: need to escape one.
        return string + "\\"

##--|

class NameParts_d:
    """A dataclass representing the parts of a person name.

    The different parts are defined according to BibTex's implementation
    of name parts (first, von, last, jr).
    """
    __slots__ = ("first", "von", "last", "jr")

    first :  list[str]
    von   :  list[str]
    last  :  list[str]
    jr    :  list[str]

    def __init__(self, *, first:MbList=None, von:MbList=None, last:MbList=None, jr:MbList=None):
        self.first = first or []
        self.von   = von   or []
        self.last  = last  or []
        self.jr    = jr    or []

    def merge(self, *, format:Maybe[list[str]]=None) -> str:
        match format:
            case None:
                format = ["von_last", "jr", "first"]
            case str():
                pass
            case x:
                raise TypeError(type(x))

        parts = {
            "first"  : " ".join(self.first) if self.first else None,
            "von"    : " ".join(self.von) if self.von else None,
            "last"   : " ".join(self.last) if self.last else None,
            "jr"     : " ".join(self.jr) if self.jr else None,
        }
        parts['von_last'] = " ".join(name for name in [parts['von'], parts['last']] if name)
        ordered = [parts[x] for x in format]
        return JOIN_STR.join(escape_last_slash(name) for name in ordered if name)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} : {self.merge()}>"



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

import bibble._interface as API
from bibtexparser import model
from bibtexparser.library import Library
from .. import DuplicateKeyHandler
# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

# Body:

class TestDuplicateHandler:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match DuplicateKeyHandler():
            case API.Middleware_p():
                assert(True)
            case x:
                raise TypeError(type(x))

    def test_good_library(self, caplog):
        obj = DuplicateKeyHandler()
        lib = Library()
        match obj.transform(lib):
            case Library():
                assert(True)
            case x:
                 assert(False), x

        assert("Duplicate Key" not in caplog.text)

    def test_bad_library_for_keys(self, caplog):
        obj = DuplicateKeyHandler()
        lib = Library()
        orig_entry =  model.Entry("test", "blah", [], 0, "this is some raw text")
        bad_entry = model.Entry("test", "blah", [], 0, "a duplicate entry key block")
        lib.add(orig_entry)
        assert(not lib.failed_blocks)
        lib.add(bad_entry)
        assert(lib.failed_blocks)
        match obj.transform(lib):
            case Library() as l2:
                assert(lib is l2)
                assert(len(l2.entries) == 2)
                assert(not l2.failed_blocks)
                assert(True)
            case x:
                 assert(False), x

        assert("Duplicate Key" in caplog.text)


    def test_bad_library_for_fields(self, caplog):
        obj = DuplicateKeyHandler()
        lib = Library()
        bad_entry = model.Entry("test", "blah",
                                [
                                    model.Field("bloo", "aweg"),
                                    model.Field("bloo", "qqqq")
                                ],
                                0, "a duplicate entry key block")
        assert("bloo_2" not in bad_entry)
        fail_block = model.DuplicateFieldKeyBlock(["bloo"], bad_entry)
        lib.add(fail_block)
        assert(lib.failed_blocks)
        match obj.transform(lib):
            case Library() as l2:
                assert(lib is l2)
                assert(len(l2.entries) == 1)
                assert(not l2.failed_blocks)
                assert(True)
            case x:
                 assert(False), x

        assert("duplicate fields" in caplog.text)
        assert("bloo" in bad_entry)
        assert("bloo_2" in bad_entry)



"""

"""

from .tags_reader import TagsReader
from .tags_writer import TagsWriter
from .key_locker import KeyLocker
from .isbn_writer import IsbnWriter
from .isbn_validator import IsbnValidator
from .metadata_writer import ApplyMetadata, FileCheck
from .entry_sorter import EntrySorter
from .data_insert import DataInsertMW



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from copy import deepcopy
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import bibtexparser
import bibtexparser.model as model
from bibtexparser import middlewares as ms
from bibtexparser.middlewares.middleware import (BlockMiddleware,
                                                 LibraryMiddleware)
from jgdv import Mixin, Proto

# ##-- end 3rd party imports

# ##-- 1st party imports
import bibble._interface as API
from . import _interface as MAPI
from bibble.util.middlecore import IdenBlockMiddleware
from bibble.util.mixins import ErrorRaiser_m
# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe, Rx, RxStr
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging
##--|

@Proto(API.ReadTime_p)
@Mixin(ErrorRaiser_m)
class KeyLocker(IdenBlockMiddleware):
    """ Ensure key/crossref consistency by:
    removing unwanted chars in the key,
    'locking' the key with a specific suffix (by default a '_').

    Also formats crossref values so they match.
    Already locked keys are ignored.

    __init__ takes:
    - regex = the regex of chars to remove.
    - sub   = the substitute for removed chars
    """

    def __init__(self, *, regex:Maybe[RxStr|Rx]=None, sub:Maybe[str]=None, lock_suffix:Maybe[str]=None, key_suffix:Maybe[RxStr|Rx]=None, **kwargs):
        super().__init__(**kwargs)
        self._remove_re         : Rx         = re.compile(regex or MAPI.KEY_CLEAN_RE)
        self._key_suffix_re     : Rx         = re.compile(key_suffix or MAPI.KEY_SUFFIX_RE)
        self._sub               : str        = sub or MAPI.KEY_SUB_CHAR
        self._lock_suffix       : str        = MAPI.LOCK_SUFFIX
        self._bad_lock          : str        = f"{self._lock_suffix}{self._lock_suffix}"

    def on_read(self):
        Never()

    def transform_Entry(self, entry, library) -> list:
        entry = deepcopy(entry)
        entry.key = self.clean_key(entry.key)
        match entry.get(MAPI.CROSSREF_K):
            case None:
                pass
            case model.Field(value=value):
                entry.set_field(model.Field(MAPI.CROSSREF_K, self.clean_key(value)))

        return [entry]

    def clean_key(self, key:str) -> str:
        """ Convert the entry key to a canonical form """
        if key.endswith(self._lock_suffix) and not key.endswith(self._bad_lock):
            return key

        # Remove bad chars
        clean_key = self._remove_re.sub(self._sub, key)
        # Enforce the correct suffix
        clean_key = self._key_suffix_re.sub(self._lock_suffix, clean_key)
        return clean_key
